import pandas as pd
import plotnine as p9

from liana._constants import Keys as K, DefaultValues as V
from liana._docs import d

@d.dedent
def target_metrics(misty,
                   stat,
                   top_n = None,
                   ascending = False,
                   key = None,
                   filterby = None,
                   filter_lambda: callable = None,
                   figure_size: tuple = (7,5),
                   return_fig: bool = V.return_fig):
    """
    Plot target metrics.

    Parameters
    ----------
    %(misty)s
    stat : str
        Statistic to plot
    %(top_n)s
    ascending : bool
        Whether to sort in ascending order
    key : callable
        Function to use to sort the dataframe
    %(filterby)s
    %(filter_lambda)s
    %(figure_size)s
    %(return_fig)s

    Returns
    -------

    Returns a plotnine plot.

    """
    target_metrics = misty.uns[K.target_metrics].copy()

    if filterby is not None:
        msk = target_metrics[filterby].apply(filter_lambda)
        target_metrics = target_metrics[msk]
    if top_n is not None:
        target_metrics = target_metrics.sort_values(stat, ascending=ascending, key=key).head(top_n)

    # get order of target by decreasing intra.R2
    targets = target_metrics.sort_values(by=stat, ascending=False)['target'].values
    # targets as categorical variable
    target_metrics['target'] = pd.Categorical(target_metrics['target'],
                                              categories=targets,
                                              ordered=True)

    p = (p9.ggplot(target_metrics, p9.aes(x='target', y=stat)) +
         p9.geom_point(size=3) +
         p9.theme_bw() +
         p9.theme(axis_text_x=p9.element_text(rotation=90),
                  figure_size=figure_size) +
        p9.labs(x='Target')
        )

    if return_fig:
        return p
    p.draw()

@d.dedent
def contributions(misty,
                  top_n=None,
                  ascending=False,
                  key=None,
                  figure_size: tuple = (7, 5),
                  return_fig: bool = V.return_fig):
    """
    Plot view contributions per target.

    Parameters
    ----------

    %(misty)s
    %(top_n)s
    ascending : bool
        Whether to sort in ascending order
    key : callable
        Function to use to sort the dataframe
    %(figure_size)s
    %(return_fig)s

    Returns
    -------
    A plotnine plot.

    """
    target_metrics = misty.uns[K.target_metrics].copy()

    view_names = misty.view_names.copy()
    if 'intra' not in target_metrics.columns:
        view_names.remove('intra')

    target_metrics = target_metrics[['target', *view_names]]
    target_metrics = target_metrics.melt(id_vars='target', var_name='view', value_name='contribution')

    if top_n is not None:
        target_metrics = target_metrics.sort_values('contribution', ascending=ascending, key=key).head(top_n)

    p = (p9.ggplot(target_metrics, p9.aes(x='target', y='contribution', fill='view')) +
            p9.geom_bar(stat='identity') +
            p9.theme_bw(base_size=14) +
            p9.theme(axis_text_x=p9.element_text(rotation=90),
                     figure_size=figure_size) +
            p9.scale_fill_brewer(palette=2, type='qual') +
            p9.labs(x='Target', y='Contribution', fill='View')
    )

    if return_fig:
        return p
    p.draw()

@d.dedent
def interactions(misty,
                 view,
                 top_n = None,
                 ascending = False,
                 key = None,
                 filterby = None,
                 filter_lambda: callable = None,
                 figure_size: tuple = (7,5),
                 return_fig: bool = V.return_fig):
    """
    Plot interaction importances.

    Parameters
    ----------

    %(misty)s
    view : str
        A view to plot
    %(top_n)s
    ascending : bool
        Whether to sort interactions in ascending order
    key : str
        Key to use when sorting interactions
    %(filterby)s
    %(filter_lambda)s
    %(figure_size)s
    %(return_fig)s

    Returns
    -------
    A plotnine plot.

    """
    interactions = misty.uns[K.interactions].copy()
    interactions = interactions[interactions['view'] == view]
    grouped = interactions.groupby('predictor')['importances'].apply(lambda x: x.isna().all())
    interactions = interactions[~interactions['predictor'].isin(grouped[grouped].index)]

    if filterby is not None:
        top_interactions = interactions[interactions[filterby].apply(filter_lambda)]
        top_interactions = top_interactions.drop_duplicates(['target', 'predictor'])
    if top_n is not None:
        interactions = interactions.sort_values(by='importances', key=key, ascending=ascending)
        top_interactions = interactions.drop_duplicates(['target', 'predictor']).head(top_n)

    if (filterby is not None) or (top_n is not None):
        interactions = interactions[interactions['target'].isin(top_interactions['target']) &
                            interactions['predictor'].isin(top_interactions['predictor'])]

    p = (p9.ggplot(interactions,
                   p9.aes(x='predictor',
                          y='target',
                          fill='importances')
                   ) +
    p9.geom_tile() +
    p9.theme_minimal(base_size=12) +
    p9.theme(axis_text_x=p9.element_text(rotation=90),
             figure_size=figure_size) +
    p9.labs(x='Predictor', y='Target', fill='Importance')
    )

    if return_fig:
        return p
    p.draw()



import numpy as np
from liana.method.sc._Method import Method, MethodMeta
from liana.method._pipe_utils._get_mean_perms import _calculate_pvals

# Internal Function to calculate CellPhoneDB LR_mean and p-values
def _cpdb_score(x, perm_stats) -> tuple:
    """
    Calculate CellPhoneDB-like LR means and p-values

    Parameters
    ----------
    x
        DataFrame with LIANA results
    perm_stats
        Permutation statistics (2 (ligand-receptor), n_perms (number of permutations, n_rows in lr_res)

    Returns
    -------
    A tuple with lr_mean and p-value for x

    """
    zero_msk = ((x['ligand_means'] == 0) | (x['receptor_means'] == 0))
    lr_means = np.mean((x['ligand_means'].values, x['receptor_means'].values), axis=0)
    lr_means[zero_msk] = 0

    cpdb_pvals = _calculate_pvals(lr_means, perm_stats, np.mean)

    return lr_means, cpdb_pvals


# Initialize CPDB Meta
_cellphonedb = MethodMeta(method_name="CellPhoneDB",
                          complex_cols=["ligand_means", "receptor_means"],
                          add_cols=[],
                          fun=_cpdb_score,
                          magnitude="lr_means",
                          magnitude_ascending=False,
                          specificity="cellphone_pvals",
                          specificity_ascending=True,
                          permute=True,
                          reference="Efremova, M., Vento-Tormo, M., Teichmann, S.A. and "
                                    "Vento-Tormo, R., 2020. CellPhoneDB: inferring cell–cell "
                                    "communication from combined expression of multi-subunit "
                                    "ligand–receptor complexes. Nature protocols, 15(4), "
                                    "pp.1484-1506. "
                          )

cellphonedb = Method(_method=_cellphonedb)



from ._sample_lrs import sample_lrs
from ._sample_anndata import generate_toy_spatial
from ._sample_anndata import generate_toy_adata
from .datasets import kang_2018
from ._sample_misty import _sample_interactions
from ._sample_misty import _sample_target_metrics
from ._sample_dea import _sample_dea




from liana.plotting import interactions, contributions, target_metrics
from liana.testing import generate_toy_spatial
from liana.testing import _sample_target_metrics, _sample_interactions

adata = generate_toy_spatial()
adata.uns['target_metrics'] = _sample_target_metrics()
adata.uns['interactions'] = _sample_interactions()
adata.view_names = ['intra', 'extra']


def test_target_contributions_plot():
    contributions(misty=adata)
    contributions(misty=adata, top_n=1, return_fig=False)


def test_target_metrics_plot():
    target_metrics(misty=adata, stat='gain_R2')
    target_metrics(misty=adata, stat='gain_R2', top_n=1, return_fig=False)
    target_metrics(misty=adata, stat='gain_R2', filterby='multi_R2', filter_lambda=lambda x: x > 0.5)


def test_misty_interactions_plot():
    interactions(misty=adata, top_n=3, view='extra', key=abs, ascending=False)



import numpy as np
import pandas as pd
from itertools import product

from liana.method._pipe_utils import prep_check_adata, assert_covered, filter_resource, \
    filter_reassemble_complexes, _check_groupby
from liana.method._pipe_utils._common import _join_stats, _get_props
from liana.method._pipe_utils._reassemble_complexes import explode_complexes
from liana.resource._select_resource import _handle_resource

from liana._docs import d
from liana._constants import DefaultValues as V, InternalValues as I, PrimaryColumns as P

@d.dedent
def df_to_lr(adata,
             dea_df,
             groupby,
             stat_keys,
             resource_name = V.resource_name,
             resource = V.resource,
             interactions = V.interactions,
             layer = V.layer,
             use_raw = V.layer,
             expr_prop = V.expr_prop,
             min_cells = V.min_cells,
             complex_col=None,
             return_all_lrs=V.return_all_lrs,
             source_labels = None,
             target_labels = None,
             lr_sep=V.lr_sep,
             verbose = V.verbose,
             ):
    """
    Convert DEA results to ligand-receptor pairs.

    Parameters
    ----------
    %(adata)s
    dea_df : pd.DataFrame
        DEA results. Index must match adata.var_names
    %(groupby)s
    stat_keys : list
        List of statistics to be used for ligand-receptor pairs
    %(resource_name)s
    %(resource)s
    %(interactions)s
    %(layer)s
    %(use_raw)s
    %(expr_prop)s
    %(min_cells)s
    complex_col : str, optional
        Column in `dea_df` to use for complex expression. Default is None.
        If None, will use mean expression ('expr') calculated per group in `groupby`.
    %(return_all_lrs)s
    %(source_labels)s
    %(target_labels)s
    %(lr_sep)s
    %(verbose)s

    Returns
    -------
    Returns a pd.DataFrame with joined ligand-receptor pairs and statistics.

    """
    _check_groupby(adata=adata, groupby=groupby, verbose=verbose)
    if (groupby not in adata.obs.columns) or (groupby not in dea_df.columns):
        raise ValueError('groupby must match a column in both adata.obs and dea_df')
    if not np.any(adata.var_names.isin(dea_df.index)):
        raise ValueError('index of dea_df must match adata.var_names')

    resource = _handle_resource(interactions=interactions,
                                resource=resource,
                                resource_name=resource_name,
                                verbose=verbose)

    stat_names = ['expr', 'props'] + stat_keys
    if complex_col is not None:
        if complex_col not in stat_names:
            raise ValueError(f'complex_col must be one of {stat_names}')
        stat_names = stat_names[stat_names.index(complex_col):]+stat_names[:stat_names.index(complex_col)]
    else:
        complex_col = 'expr'

    # Check and Reformat Mat if needed
    adata = prep_check_adata(adata=adata,
                             groupby=groupby,
                             min_cells=min_cells,
                             use_raw=use_raw,
                             layer=layer,
                             verbose=verbose
                             )

    # reduce dim of adata
    adata =  adata[:, dea_df.index.unique()]

    # get label cats
    labels = adata.obs[I.label].cat.categories

    dedict = {}

    for label in labels:
        temp = adata[adata.obs[I.label] == label, :]
        props = _get_props(temp.X)
        means = np.array(temp.X.mean(axis=0), dtype='float32').flatten()

        stats = pd.DataFrame({'names': temp.var_names,
                              'props': props,
                              'expr': means
                              }). \
            assign(label=label).sort_values('names')

        # merge DEA results to props & means
        dea_df.index.name = None
        df = dea_df[dea_df[groupby] == label].drop(groupby, axis=1). \
            reset_index().rename(columns={'index': "names"})

        if not return_all_lrs:
            stats = stats.merge(df, on='names')
        else:
            stats = df.merge(stats, on='names', how='outer')

        dedict[label] = stats[['names', 'label', *stat_names]]
        all_stats = pd.concat(dedict.values())

    # Create df /w cell identity pairs
    pairs = pd.DataFrame(list(product(labels, labels))). \
        rename(columns={0: "source", 1: "target"})

    if source_labels is not None:
        pairs = pairs[pairs['source'].isin(source_labels)]
    if target_labels is not None:
        pairs = pairs[pairs['target'].isin(target_labels)]

    resource = explode_complexes(resource)

    # Check overlap between resource and adata
    assert_covered(np.union1d(np.unique(resource["ligand"]),
                                np.unique(resource["receptor"])),
                    all_stats['names'], verbose=verbose)

    # Filter Resource
    resource = filter_resource(resource, all_stats['names'].unique())

    # Join Stats to LR
    lr_res = pd.concat(
        [_join_stats(source, target, dedict, resource) for
            source, target in zip(pairs['source'], pairs['target'])]
    )

    # ligand_ or receptor + stat_keys
    complex_cols = list(product(['ligand', 'receptor'], [complex_col]))
    complex_cols = [f'{x}_{y}' for x, y in complex_cols]

    # assign receptor and ligand absolutes, NOTE deals with missing values
    _placeholders = ['ligand_absolute', 'receptor_absolute']
    lr_res[_placeholders] = \
        lr_res[complex_cols].apply(lambda x: x.abs())
    if return_all_lrs:
        lr_res[_placeholders] = lr_res[_placeholders].fillna(0)

    lr_res = filter_reassemble_complexes(lr_res=lr_res,
                                         _key_cols=P.primary,
                                         expr_prop=expr_prop,
                                         return_all_lrs=return_all_lrs,
                                         complex_cols=_placeholders
                                        )
    lr_res = lr_res.drop(['prop_min', 'interaction', *_placeholders], axis=1)

    # summarise stats for each lr
    for key in stat_keys:
        stat_columns = lr_res.columns[lr_res.columns.str.endswith(key)]
        lr_res.loc[:, f'interaction_{key}'] = lr_res.loc[:, stat_columns].mean(axis=1)

    lr_res['interaction'] = lr_res['ligand_complex'] + lr_sep + lr_res['receptor_complex']

    return lr_res



from __future__ import annotations

import pandas as pd
from liana._docs import d
from anndata import AnnData
from mudata import MuData

@d.dedent
def get_factor_scores(adata:AnnData | MuData,
                      obsm_key: str = None,
                      obs_keys: str | None = None):
    """
    Extract factor scores from an AnnData object.

    Parameters
    ----------
    %(adata)s
    obsm_key: str
        Key to use when extracting factor scores from `adata.obsm`
    obs_keys: list
        List of keys to use when extracting metadata from `adata.obs`
        If None, no metadata is extracted. Default is None.

    Returns
    -------

    Returns a pandas DataFrame with the factor scores.

    """
    if obsm_key not in adata.obsm.keys():
        raise ValueError(f'{obsm_key} not found in `.obsm`')

    df = pd.DataFrame(adata.obsm[obsm_key], index=adata.obs.index)

    df.columns = ['Factor{0}'.format(x + 1) for x in range(df.shape[1])]
    df = df.reset_index()

    # join with metadata
    if obs_keys is not None:
        obs = adata.obs[obs_keys].reset_index()
        df = df.merge(obs)

    return df

@d.dedent
def get_variable_loadings(adata: AnnData | MuData,
                          varm_key:str = None,
                          view_sep:str | None = None,
                          variable_sep:str | None = None,
                          pair_sep:str | None = None,
                          var_names:list = ['ligand_complex', 'receptor_complex'],
                          pair_names:list = ['source', 'target'],
                          drop_columns:bool = True
                          ):
    """
    Extract variable loadings from an AnnData object.

    Parameters
    ----------

    %(adata)s
    varm_key: str
        Key to use when extracting variable loadings from `mdata.varm`
    view_sep: str
        Separator to use when splitting view:variable names into view and variable
    variable_sep: str
        Separator to use when splitting variable names into `var_names` ('ligand_complex' and 'receptor_complex' by default)
    pair_sep: str
        Separator to use when splitting view names into `pair_names` ('source' and 'target' by default)
    drop_columns: bool
        If True, drop the `view:variable` column

    Returns
    -------

    Returns a pandas DataFrame with the variable loadings for the specified index.
    """
    if varm_key not in adata.varm.keys():
        raise ValueError(f'{varm_key} not found in adata.varm')

    n_factors = adata.varm[varm_key].shape[1]
    columns = [f'Factor{i+1}' for i in range(n_factors)]

    df = pd.DataFrame(index=adata.var.index, data=adata.varm[varm_key], columns=columns)

    df.index.name = None
    df = df.reset_index()

    if view_sep:
        df[['view', 'variable']] = df['index'].str.split(view_sep, expand=True)

        if drop_columns:
            df.drop(columns='index', inplace=True)

    if variable_sep:
        if view_sep is None:
            df = df.rename(columns={'index': 'variable'})

        df[var_names] = df['variable'].str.split(variable_sep, expand=True)

        if drop_columns:
            df.drop(columns='variable', inplace=True)

    if pair_sep:
        df[pair_names] = df['view'].str.split(pair_sep, expand=True)

        if drop_columns:
            df.drop(columns='view', inplace=True)

    # Re-order columns so that factors are last
    df = df.reindex(sorted(df.columns, key=lambda x: x.startswith('Factor')), axis=1)

    # re-order to absolute values
    df = (df.reindex(df['Factor1'].abs().sort_values(ascending=False).index))

    return df



from plotnine import ggplot, aes, geom_point, theme_minimal, labs, theme
import anndata
import pandas as pd

from liana._logging import _logg
from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V

@d.dedent
def connectivity(adata: anndata.AnnData,
                 idx: int,
                 spatial_key=K.spatial_key,
                 connectivity_key=K.connectivity_key,
                 size=1,
                 figure_size=(5.4, 5),
                 return_fig: bool = V.return_fig):
    """
    Plot spatial connectivity weights.

    Parameters
    ----------
    %(adata)s
    %(spatial_key)s
    %(connectivity_key)s
    size
        Size of the points
    %(figure_size)s
    %(return_fig)s

    Returns
    -------
    A `plotnine.ggplot` instance

    """

    assert connectivity_key in list(adata.obsp.keys())
    assert spatial_key in adata.obsm_keys()

    _logg("This function will be deprecated in the next version. " +
          "Please use scanpy or squidpy for plotting spatial connectivities.", level='warn')

    coordinates = pd.DataFrame(adata.obsm[spatial_key],
                               index=adata.obs_names,
                               columns=['x', 'y']).copy()
    coordinates['connectivity'] = adata.obsp[connectivity_key][:, idx].A
    coordinates['y'] = coordinates['y'].max() - coordinates['y'] # flip y

    p = (ggplot(coordinates.sort_values('connectivity', ascending=True),
                aes(x='x', y='y', colour='connectivity'))
         + geom_point(size=size, shape='8')
         + theme_minimal()
         + labs(colour='connectivity', y='y Coordinate', x='x Coordinate')
         + theme(figure_size=figure_size)
         )


    if return_fig:
        return p

    p.draw()



import numpy as np
from liana.testing._sample_anndata import generate_toy_spatial

from liana.method.sp._lr_bivar import lr_bivar

adata = generate_toy_spatial()


def test_morans_analytical():
    adata = generate_toy_spatial()
    lr_bivar(adata, function_name='morans', n_perms=0, use_raw=True, mask_negatives=True)
    assert 'local_scores' in adata.obsm_keys()
    lrdata = adata.obsm['local_scores']

    assert 'global_res' in lrdata.uns_keys()
    assert 'pvals' in lrdata.layers.keys()

    # test specific interaction
    global_res = lrdata.uns['global_res']
    interaction = global_res[global_res.interaction == 'S100A9^ITGB2']
    np.testing.assert_almost_equal(interaction['global_r'].values, 0.0994394)
    np.testing.assert_almost_equal(interaction['global_pvals'].values, 3.4125671e-07)

    np.testing.assert_almost_equal(np.mean(lrdata[:,'MIF^CD74_CXCR4'].X), -0.008322, decimal=6)
    np.testing.assert_almost_equal(np.mean(lrdata[:,'MIF^CD74_CXCR4'].layers['pvals']), 0.6546573748814639, decimal=6)


def test_cosine_permutation():
    adata = generate_toy_spatial()
    lr_bivar(adata, function_name='cosine', n_perms=100, use_raw=True)
    lrdata = adata.obsm['local_scores']

    assert 'global_res' in lrdata.uns_keys()
    assert 'pvals' in lrdata.layers.keys()

    global_res = adata.uns['global_res']
    interaction = global_res[global_res.interaction == 'S100A9^ITGB2']

    np.testing.assert_almost_equal(interaction['global_mean'].values, 0.56016606)
    np.testing.assert_almost_equal(interaction['global_sd'].values, 0.33243373)

    np.testing.assert_almost_equal(lrdata[:,'MIF^CD74_CXCR4'].X.mean(), 0.32514292, decimal=6)
    np.testing.assert_almost_equal(np.mean(lrdata[:,'MIF^CD74_CXCR4'].layers['pvals']), 0.6274714285714286, decimal=4)


def test_morans_pval_none_cats():
    adata = generate_toy_spatial()
    lr_bivar(adata, function_name='morans', n_perms=None, use_raw=True, add_categories=True)
    assert 'global_res' in adata.uns_keys()
    assert 'local_scores' in adata.obsm_keys()
    # NOT IN
    adata = adata.obsm['local_scores']
    assert 'pvals' not in adata.layers.keys()

    assert 'cats' in adata.layers.keys()
    assert adata.layers['cats'].sum() == -6197

def test_wrong_interactions():
    from pytest import raises
    with raises(ValueError):
        lr_bivar(adata, resource_name='mouseconsensus', function_name='morans', n_perms=None, use_raw=True, add_categories=True)



from __future__ import annotations

from anndata import AnnData
import pandas as pd
from typing import Optional

from liana.method.sp._SpatialBivariate import SpatialBivariate
from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V

class SpatialLR(SpatialBivariate):
    """ A child class of SpatialBivariate for ligand-receptor analysis. """
    def __init__(self):
        super().__init__(x_name='ligand', y_name='receptor')

    @d.dedent
    def __call__(self,
                 adata: AnnData,
                 function_name: str,
                 resource_name: str = V.resource_name,
                 resource: Optional[pd.DataFrame] = V.resource,
                 interactions: list = V.interactions,
                 expr_prop: float = 0.05,
                 n_perms: (None | int) = None,
                 mask_negatives: bool = False,
                 seed: int = V.seed,
                 add_categories: bool = False,
                 use_raw: Optional[bool] = V.use_raw,
                 layer: Optional[str] = V.layer,
                 connectivity_key = K.connectivity_key,
                 inplace = True,
                 key_added='global_res',
                 obsm_added='local_scores',
                 lr_sep=V.lr_sep,
                 verbose: Optional[bool] = V.verbose,
                 ):
        """
        Local ligand-receptor interaction metrics and global scores.

        Parameters
        ----------
        %(adata)s
        %(function_name)s
        %(resource_name)s
        %(resource)s
        %(interactions)s
        %(expr_prop)s
        %(n_perms)s
        %(mask_negatives)s
        %(seed)s
        %(add_categories)s
        %(use_raw)s
        %(layer)s
        %(connectivity_key)s
        %(inplace)s
        %(key_added)s
        obsm_added: str
            Key in `adata.obsm` where the local scores are stored.
        %(lr_sep)s
        %(verbose)s

        Returns
        -------
        Returns `adata` with the following fields, if `inplace=True`:
            - `adata.uns[key_added]` - global results (pd.DataFrame)
            - `adata.obsm[obsm_added]` - local scores (AnnData object)
        if `inplace=False`, returns a the above.
        """

        lr_res, local_scores = super().__call__(
            mdata=adata,
            function_name=function_name,
            connectivity_key=connectivity_key,
            resource_name=resource_name,
            resource=resource,
            interactions=interactions,
            nz_threshold=expr_prop,
            n_perms=n_perms,
            mask_negatives=mask_negatives,
            add_categories=add_categories,
            x_mod=True,
            y_mod=True,
            x_use_raw=use_raw,
            x_layer=layer,
            seed=seed,
            verbose=verbose,
            xy_sep=lr_sep,
            inplace=False,
            complex_sep='_'
            )

        return self._handle_return(adata, lr_res, local_scores, key_added, obsm_added, inplace)

lr_bivar = SpatialLR()



from __future__ import annotations

import anndata
import numpy as np
from tqdm import tqdm


def _get_means_perms(adata: anndata.AnnData,
                     n_perms: int,
                     seed: int,
                     agg_fun,
                     norm_factor: float | None,
                     verbose: bool):
    """
    Generate permutations and indices required for permutation-based methods

    Parameters
    ----------
    adata
        Annotated data matrix
    n_perms
        Number of permutations to be calculated
    seed
        Random seed for reproducibility.
    agg_fun
        function by which to aggregate the matrix, should take `axis` argument
    norm_factor
        additionally normalize the data by some factor (e.g. matrix max for CellChat)
    verbose
        Verbosity bool

    Returns
    -------
    Tuple with:
        - perms: 3D tensor with permuted averages per cluster
        - ligand_pos: Index of the ligand in the tensor
        - receptor_pos: Index of the receptor in the perms tensor
        - labels_pos: Index of cell identities in the perms tensor
    """

    if isinstance(norm_factor, np.float32):
        adata.X /= norm_factor

    # define labels and masks
    labels = adata.obs['@label'].cat.categories
    labels_mask = np.zeros((adata.shape[0], labels.shape[0]), dtype=bool)

    # populate masks shape(genes, labels)
    for ct_idx, label in enumerate(labels):
        labels_mask[:, ct_idx] = adata.obs['@label'] == label

    # Perm should be a cube /w dims: n_perms x idents x n_genes
    perms = _generate_perms_cube(adata.X, n_perms, labels_mask, seed, agg_fun, verbose)

    return perms


# @nb.njit(nb.float32[:,:,:](nb.float32[:,:], nb.int32[:,:], nb.float32[:]), cache=True)
def _generate_perms_cube(X, n_perms, labels_mask, seed, agg_fun, verbose):
    # initialize rng
    rng = np.random.default_rng(seed=seed)

    # indexes to be shuffled
    idx = np.arange(X.shape[0])

    # Perm should be a cube /w dims: n_perms x idents x n_genes
    perms = np.zeros((n_perms, labels_mask.shape[1], X.shape[1])) # , dtype=np.float32

    # Assign permuted matrix
    for perm in tqdm(range(n_perms), disable=not verbose):
        perm_idx = rng.permutation(idx)
        perm_mat = X[perm_idx]
        # populate matrix /w permuted means
        for ct_idx in range(labels_mask.shape[1]):
            ct_mask = labels_mask[:, ct_idx]
            perms[perm, ct_idx] = agg_fun(perm_mat[ct_mask], axis=0)

    return perms


def _get_positions(adata, lr_res):
    labels = adata.obs['@label'].cat.categories

    # get positions of each entity in the matrix
    ligand_pos = {entity: np.where(adata.var_names == entity)[0][0] for entity
                  in lr_res['ligand']}
    receptor_pos = {entity: np.where(adata.var_names == entity)[0][0] for entity
                    in lr_res['receptor']}
    labels_pos = {labels[pos]: pos for pos in range(labels.shape[0])}

    return ligand_pos, receptor_pos, labels_pos


def _get_mat_idx(adata, lr_res):
    # convert to indexes
    ligand_pos, receptor_pos, labels_pos = _get_positions(adata, lr_res)

    ligand_idx = lr_res['ligand'].map(ligand_pos)
    receptor_idx = lr_res['receptor'].map(receptor_pos)

    source_idx = lr_res['source'].map(labels_pos)
    target_idx = lr_res['target'].map(labels_pos)

    return ligand_idx, receptor_idx, source_idx, target_idx



def _calculate_pvals(lr_truth, perm_stats, _score_fun):
    """
    Calculate p-values for a given DataFrame x and permutation statistics

    Parameters
    ----------
    x
        DataFrame with LIANA results
    perm_stats
        Permutation statistics (2 (ligand-receptor), n_perms (number of permutations, n_rows in lr_res)

    Returns
    -------
    A tuple with lr_mean and pvalue for x

    """
    # calculate p-values
    if perm_stats is not None:
        lr_perm_means = _score_fun(perm_stats, axis=0)
        n_perms = perm_stats.shape[1]
        pvals = np.sum(np.greater_equal(lr_perm_means, lr_truth), axis=0) / n_perms
    else:
        pvals = None


    return pvals



import numba as nb
import numpy as np
from scipy.stats import rankdata


@nb.njit(nb.float32(nb.float32[:], nb.float32[:], nb.float32[:], nb.float32), cache=True)
def _wcorr(x, y, w, wsum):

    x = np.argsort(x).argsort().astype(nb.float32)
    y = np.argsort(y).argsort().astype(nb.float32)

    wx = w * x
    wy = w * y

    numerator = wsum * sum(wx * y) - sum(wx) * sum(wy)

    denominator_x = wsum * sum(w * (x**2)) - sum(wx)**2
    denominator_y = wsum * sum(w * (y**2)) - sum(wy)**2
    denominator = (denominator_x * denominator_y)

    if (denominator == 0) or (numerator == 0):
        return 0

    return numerator / (denominator**0.5)


@nb.njit(nb.float32[:,:](nb.float32[:,:], nb.float32[:,:], nb.float32[:,:]), parallel=True, cache=True)
def _masked_spearman(x_mat, y_mat, weight):
    spot_n = x_mat.shape[0]
    xy_n = x_mat.shape[1]

    local_corrs = np.zeros((spot_n, xy_n), dtype=nb.float32)

    for i in nb.prange(spot_n):
        w = weight[i, :]
        msk = w > 0
        wsum = sum(w[msk])

        for j in range(xy_n):
            x = x_mat[:, j][msk]
            y = y_mat[:, j][msk]

            local_corrs[i, j] = _wcorr(x, y, w[msk], wsum)

    # NOTE done due to numpy/numba sum imprecision, https://github.com/numba/numba/issues/8749
    local_corrs = np.clip(a=local_corrs, a_min=-1.0, a_max=1.0, out=local_corrs)

    return local_corrs.T


def _vectorized_correlations(x_mat, y_mat, weight, method="pearson"):
    """
    Vectorized implementation of weighted correlations.

    Note: due to the imprecision of np.sum and np.dot, the function is accurate to 5 decimal places.

    """
    if method not in ["pearson", "spearman"]:
        raise ValueError("method must be one of 'pearson', 'spearman'")

    # transpose
    x_mat, y_mat, weight = x_mat.T, y_mat.T, weight.T

    weight_sums = np.array(np.sum(weight, axis = 0)).flatten()

    if method=="spearman":
        x_mat = rankdata(x_mat, axis=1)
        y_mat = rankdata(y_mat, axis=1)

    # standard pearson
    n1 = ((x_mat * y_mat) @ weight) * weight_sums
    n2 = (x_mat @ weight) * (y_mat @ weight)
    numerator = n1 - n2

    denominator_x = (weight_sums * (x_mat ** 2 @ weight)) - (x_mat @ weight)**2
    denominator_y = (weight_sums * (y_mat ** 2 @ weight)) - (y_mat @ weight)**2
    denominator = (denominator_x * denominator_y)

    # numpy sum is unstable below 1e-6...
    denominator[denominator < 1e-6] = 0
    denominator = denominator ** 0.5

    zeros = np.zeros(numerator.shape)
    local_corrs = np.divide(numerator, denominator, out=zeros, where=denominator!=0)

    # NOTE done due to numpy/numba sum imprecision, https://github.com/numba/numba/issues/8749
    local_corrs = np.clip(local_corrs, -1, 1, out=local_corrs, dtype=np.float32)

    return local_corrs


def _vectorized_pearson(x_mat, y_mat, weight):
    return _vectorized_correlations(x_mat, y_mat, weight, method="pearson")


def _vectorized_spearman(x_mat, y_mat, weight):
    return _vectorized_correlations(x_mat, y_mat, weight, method="spearman")


def _vectorized_cosine(x_mat, y_mat, weight):

    x_mat, y_mat = x_mat.T, y_mat.T

    xy_dot = (x_mat * y_mat) @ weight.T
    x_dot = ((x_mat ** 2) @ weight.T)
    y_dot = ((y_mat ** 2) @ weight.T)
    denominator = (x_dot * y_dot) + np.finfo(np.float32).eps

    return xy_dot / denominator**0.5


def _vectorized_jaccard(x_mat, y_mat, weight):
    # binarize
    x_mat, y_mat = x_mat > 0, y_mat > 0 ## TODO, only positive?
    # transpose
    x_mat, y_mat, weight = x_mat.T, y_mat.T, weight.T

    # intersect and union
    numerator = np.minimum(x_mat, y_mat) @ weight
    denominator = np.maximum(x_mat, y_mat) @ weight + np.finfo(np.float32).eps

    return numerator / denominator


def _local_morans(x_mat, y_mat, weight):
    """

    Parameters
    ----------
    x_mat
        2D array with x variables
    y_mat
        2D array with y variables

    Returns
    -------
    Returns 2D array of local Moran's I with shape(n_spot, xy_n)

    """

    local_x = x_mat * (weight @ y_mat)
    local_y = y_mat * (weight @ x_mat)
    local_r = (local_x + local_y).T

    return local_r


def _product(x_mat, y_mat, weight):
    x_mat = (weight @ x_mat)
    y_mat = (weight @ y_mat)
    score = x_mat * y_mat

    return score.T


def _norm_product(x_mat, y_mat, weight):

    x_mat = (weight @ x_mat)
    y_mat = (weight @ y_mat)

    x_norm = np.max(np.abs(x_mat), axis=0)
    y_norm = np.max(np.abs(y_mat), axis=0)

    x_norm[x_norm == 0.] = 1.
    y_norm[y_norm == 0.] = 1.

    x_mat = x_mat / x_norm
    y_mat = y_mat / y_norm

    score = x_mat * y_mat

    return score.T


class SpatialFunction:
    """
    Class representing information about bivariate spatial functions.
    """
    def __init__(self, name, metadata, local_function, reference=None):
        self.name = name
        self.metadata = metadata
        self.local_function = local_function
        self.reference = reference

    def __repr__(self):
        return f"{self.name}: {self.metadata}"


_bivariate_functions = [
        SpatialFunction(
            name="pearson",
            metadata="weighted Pearson correlation coefficient",
            local_function = _vectorized_pearson,
        ),
        SpatialFunction(
            name="spearman",
            metadata="weighted Spearman correlation coefficient",
            local_function = _vectorized_spearman,
        ),
        SpatialFunction(
            name="cosine",
            metadata="weighted Cosine similarity",
            local_function = _vectorized_cosine,
        ),
        SpatialFunction(
            name="jaccard",
            metadata="weighted Jaccard similarity",
            local_function = _vectorized_jaccard,
        ),
        SpatialFunction(
            name="product",
            metadata="simple weighted product",
            local_function = _product,
        ),
        SpatialFunction(
            name="norm_product",
            metadata="normalized weighted product",
            local_function = _norm_product,
        ),
        SpatialFunction(
            name="morans",
            metadata="Moran's R",
            local_function=_local_morans,
            reference="Li, Z., Wang, T., Liu, P. and Huang, Y., 2022. SpatialDM:"
            "Rapid identification of spatially co-expressed ligand-receptor"
            "reveals cell-cell communication patterns. bioRxiv, pp.2022-08."
        ),
        SpatialFunction(
            name= "masked_spearman",
            metadata="masked & weighted Spearman correlation",
            local_function=_masked_spearman,
            reference="Ghazanfar, S., Lin, Y., Su, X., Lin, D.M., Patrick, E., Han, Z.G., Marioni, J.C. and Yang, J.Y.H., 2020."
            "Investigating higher-order interactions in single-cell data with scHOT. Nature methods, 17(8), pp.799-806."
        ),
    ]


def _get_method_names():
    return [function.name for function in _bivariate_functions]

def _handle_functions(method_name):
    method_name = method_name.lower()
    for function in _bivariate_functions:
        if function.name == method_name:
            return function.local_function
    raise ValueError("The function is not implemented."
                     "Implemented functions are: {}".format(_get_method_names()))



"""
Preprocessing functions.
Functions to preprocess the anndata object prior to running any method.
"""
from __future__ import annotations

import numpy as np
from anndata import AnnData
from typing import Optional
from pandas import DataFrame, Index
import scanpy as sc
from scipy.sparse import csr_matrix, isspmatrix_csr
from liana._logging import _logg

def assert_covered(
        subset,
        superset,
        subset_name: str = "resource",
        superset_name: str = "var_names",
        prop_missing_allowed: float = 0.98,
        verbose: bool = False) -> None:
    """
    Assert if elements are covered at a decent proportion

    Parameters
    ----------
    subset
        Subset of elements
    superset
        The superset of elements
    subset_name
        Name of the subset
    superset_name
        Name of the superset
    prop_missing_allowed
        Allowed proportion of missing/mismatched elements in the subset
    verbose
        Verbosity flag

    Returns
    -------
    None
    """

    subset = np.asarray(subset)
    is_missing = ~np.isin(subset, superset)
    if subset.size == 0:
        prop_missing = 1.
        x_missing = 'values in interactions argument'
    else:
        prop_missing = np.sum(is_missing) / len(subset)
        x_missing = ", ".join([x for x in subset[is_missing]])
    if prop_missing > prop_missing_allowed:
        msg = (
            f"Please check if appropriate organism/ID type was provided! "
            f"Allowed proportion ({prop_missing_allowed}) of missing "
            f"{subset_name} elements exceeded ({prop_missing:.2f}). "
            f"Too few features from the resource were found in the data."
        )
        raise ValueError(msg + f" [{x_missing}] missing from {superset_name}")

    _logg(f"{prop_missing:.2f} of entities in the resource are missing from the data.", verbose=verbose & (prop_missing > 0))


def prep_check_adata(adata: AnnData,
                     groupby: (str | None),
                     min_cells: (int | None),
                     use_raw: Optional[bool] = False,
                     layer: Optional[str] = None,
                     obsm = None,
                     complex_sep='_',
                     verbose: Optional[bool] = False) -> AnnData:
    """
    Check if the anndata object is in the correct format and preprocess

    Parameters
    ----------
    adata
        Un-formatted Anndata.
    groupby
        column to groupby. None if the ligand-receptor pipe
        calling this function does not rely on cell labels.
        For example, if ligand-receptor stats are needed
        for the whole sample (global).
    min_cells
        minimum cells per cell identity. None if groupby is not passed.
    use_raw
        Use raw attribute of adata if present.
    layer
        Indicate whether to use any layer.
    verbose
        Verbosity flag.

    Returns
    -------
    Anndata object to be used downstream
    """
    X = _choose_mtx_rep(adata=adata, use_raw=use_raw,
                        layer=layer, verbose=verbose)

    if use_raw & (layer is None):
        var = DataFrame(index=adata.raw.var_names)
    else:
        var = DataFrame(index=adata.var_names)

    adata = sc.AnnData(X=X,
                       obs=adata.obs.copy(),
                       dtype="float32",
                       var=var,
                       obsp=adata.obsp.copy(),
                       obsm=obsm
                       ).copy()
    adata.var_names_make_unique()

    # Check for empty features
    msk_features = np.sum(adata.X, axis=0).A1 == 0
    n_empty_features = np.sum(msk_features)
    if n_empty_features > 0:
        _logg(f"{n_empty_features} features of mat are empty, they will be removed.", level='warn', verbose=verbose)
        adata = adata[:, ~msk_features]

    # Check for empty samples
    msk_samples = adata.X.sum(axis=1).A1 == 0
    n_empty_samples = np.sum(msk_samples)
    if n_empty_samples > 0:
        _logg(f"{n_empty_samples} samples of mat are empty, they will be removed.", level='warn', verbose=verbose)

    # Check if log-norm
    _sum = np.sum(adata.X.data[0:100])
    if _sum == np.floor(_sum):
        _logg("Make sure that normalized counts are passed!", level='warn', verbose=verbose)

    # Check for non-finite values
    if np.any(~np.isfinite(adata.X.data)):
        raise ValueError("mat contains non finite values (nan or inf), please set them to 0 or remove them.")

    if groupby is not None:
        _check_groupby(adata, groupby, verbose)
        adata.obs.loc[:, '@label'] = adata.obs[groupby]

        # Remove any cell types below X number of cells per cell type
        count_cells = adata.obs.groupby(groupby)[groupby].size().reset_index(name='count').copy()
        count_cells['keep'] = count_cells['count'] >= min_cells

        if not all(count_cells.keep):
            lowly_abundant_idents = list(count_cells[~count_cells.keep][groupby])
            # remove lowly abundant identities
            msk = ~np.isin(adata.obs[[groupby]], lowly_abundant_idents)
            adata = adata[msk]
            _logg("The following cell identities were excluded: {0}".format(", ".join(lowly_abundant_idents)),
                 level='warn', verbose=verbose)


    check_vars(adata.var_names,
               complex_sep=complex_sep,
               verbose=verbose)

    # Re-order adata vars alphabetically
    adata = adata[:, np.sort(adata.var_names)]

    return adata


# format variable names
def check_vars(var_names, complex_sep, verbose=False) -> list:
    """
    Raise a warning if `complex_sep` is part of any variable name.
    """
    var_issues = []
    if complex_sep is not None:
        for name in var_names:
            if complex_sep in name:
                var_issues.append(name)
    else:
        pass

    _logg(f"{var_issues} contain `{complex_sep}`. Consider replacing those!",
         verbose=verbose & (len(var_issues) > 0), level='warn')



def filter_resource(resource: DataFrame, var_names: Index) -> DataFrame:
    """
    Filter interactions for which vars are not present.

    Note that here I remove any interaction that /w genes that are not found
    in the dataset. Note that this is not necessarily the case in liana-r.
    There, I assign the expression of those with missing subunits to 0, while
    those without any subunit present are implicitly filtered.

    Parameters
    ---------
    resource
        Resource with 'ligand' and 'receptor' columns
    var_names
        Relevant variables - i.e. the variables to be used downstream

    Returns
    ------
    A filtered resource dataframe
    """
    # Remove those without any subunit
    resource = resource[(np.isin(resource.ligand, var_names)) &
                        (np.isin(resource.receptor, var_names))]

    # Only keep interactions /w complexes for which all subunits are present
    missing_comps = resource[resource.interaction.str.contains('_')].copy()
    missing_comps['all_units'] = \
        missing_comps['ligand_complex'] + '_' + missing_comps[
            'receptor_complex']

    # Get those not with all subunits
    missing_comps = missing_comps[np.logical_not(
        [all([x in var_names for x in entity.split('_')])
         for entity in missing_comps.all_units]
    )]
    # Filter them
    return resource[~resource.interaction.isin(missing_comps.interaction)]


def _choose_mtx_rep(adata, use_raw=False, layer=None, verbose=False) -> csr_matrix:
    """
    Choose matrix (adapted from scanpy)

    Parameters
    ----------
    adata
        Unformatted Anndata.
    use_raw
        Use raw attribute of adata if present.
    layer
        Indicate whether to use any layer.

    Returns
    -------
        The matrix to be used by liana-py.
    """
    is_layer = layer is not None
    if is_layer & use_raw:
        raise ValueError("Cannot specify `layer` and have `use_raw=True`.")
    if is_layer:
        _logg(f"Using the `{layer}` layer!", verbose=verbose)
        X = adata.layers[layer]
    elif use_raw:
        if adata.raw is None:
            raise ValueError("`.raw` is not initialized!")
        _logg("Using `.raw`!", verbose=verbose)
        X = adata.raw.X
    else:
        _logg("Using `.X`!", verbose=verbose)
        X = adata.X

    # convert to sparse csr matrix
    if not isspmatrix_csr(X):
        _logg("Converting to sparse csr matrix!", verbose=verbose)
        X = csr_matrix(X)

    return X

def _check_groupby(adata, groupby, verbose):
    if groupby not in adata.obs.columns:
        raise AssertionError(f"`{groupby}` not found in `adata.obs.columns`.")
    if not adata.obs[groupby].dtype.name == 'category':
        _logg(f"Converting `{groupby}` to categorical!", level='warn', verbose=verbose)
        adata.obs[groupby] = adata.obs[groupby].astype('category')



import numpy as np
from  scipy.sparse import csr_matrix

from liana.method.sp._spatial_pipe import _global_zscore_pvals, _global_permutation_pvals, _local_permutation_pvals, _local_zscore_pvals
from liana.method.sp._bivariate_funs import _local_morans
from liana.utils.spatial_neighbors import spatial_neighbors

from liana.testing._sample_anndata import generate_toy_spatial

adata = generate_toy_spatial()

def test_get_spatial_connectivities():
    spatial_neighbors(adata=adata, bandwidth=200, set_diag=True, cutoff=0.2)
    np.testing.assert_equal(adata.obsp['spatial_connectivities'].shape, (adata.shape[0], adata.shape[0]))
    np.testing.assert_equal(adata.obsp['spatial_connectivities'].sum(), 4550.654013895928)

    spatial_neighbors(adata=adata, bandwidth=100, set_diag=True, cutoff=0.1)
    np.testing.assert_equal(adata.obsp['spatial_connectivities'].sum(), 1802.332962418902)

    conns = spatial_neighbors(adata=adata, bandwidth=100,
                              kernel='linear', cutoff=0.1,
                              set_diag=True, inplace=False)
    assert conns.sum() == 899.065036633088

    conns = spatial_neighbors(adata=adata, bandwidth=100,
                              kernel='exponential', cutoff=0.1,
                              set_diag=True, inplace=False)
    assert conns.sum() == 1520.8496098963612

    conns = spatial_neighbors(adata=adata, bandwidth=100, set_diag=True,
                              kernel='misty_rbf', cutoff=0.1,
                              inplace=False)
    assert conns.sum() == 1254.3161716188595

    conns = spatial_neighbors(adata=adata, bandwidth=250, set_diag=False,
                              max_neighbours=100,
                              kernel='gaussian', cutoff=0.1,
                              inplace=False)
    assert conns.sum() == 6597.05237692107

    conns = spatial_neighbors(adata=adata, bandwidth=250,
                              set_diag=False, max_neighbours=100,
                              kernel='gaussian', cutoff=0.1,
                              inplace=False, standardize=True)
    np.testing.assert_almost_equal(conns.sum(), conns.shape[0])

# toy test data
seed = 0
rng = np.random.default_rng(seed=seed)
dist = csr_matrix(rng.normal(size=(10, 10)))

norm_factor = dist.shape[0] / dist.sum()
weight = csr_matrix(norm_factor * dist)

x_mat = rng.normal(size=(10, 10))
y_mat = rng.normal(size=(10, 10))
n_perms = 100
mask_negatives = True


def test_global_permutation_pvals():
    global_truth = rng.normal(size=(10))

    pvals = _global_permutation_pvals(x_mat=x_mat,
                                      y_mat=y_mat,
                                      global_r=global_truth,
                                      seed=seed,
                                      n_perms=n_perms,
                                      mask_negatives=mask_negatives,
                                      weight = weight,
                                      verbose=False
                                      )
    assert pvals.shape == (10, )
    assert pvals.sum().round(3)==4.65



def test_local_permutation_pvals():
    local_truth = rng.normal(size=(10, 10))
    mask_negatives = True

    pvals = _local_permutation_pvals(x_mat = x_mat,
                                     y_mat = y_mat,
                                     local_truth = local_truth,
                                     local_fun = _local_morans,
                                     weight = weight,
                                     n_perms = n_perms,
                                     seed = seed,
                                     mask_negatives=mask_negatives,
                                     local_msk=np.ones((10, 10), dtype=bool),
                                     verbose=False
                                     )
    assert pvals.shape == (10, 10)


def test_global_zscore_pvals():
    global_truth = rng.normal(size=(10))
    pvals = _global_zscore_pvals(global_r=global_truth,
                                 weight=weight,
                                 mask_negatives=mask_negatives
    )
    assert pvals.shape == (10, )


def test_local_zscore_pvals():
    local_truth = rng.normal(size=(10, 10))
    pvals = _local_zscore_pvals(x_mat=x_mat, y_mat=y_mat, weight=weight,
                                local_truth=local_truth, mask_negatives=mask_negatives,
                                local_msk=np.ones((10, 10), dtype=bool))
    assert pvals.shape == (10, 10)



from __future__ import annotations

from functools import reduce

import anndata
import pandas
import scanpy as sc
import pandas as pd
import numpy as np

from liana.method._pipe_utils import prep_check_adata, assert_covered, \
    filter_resource, filter_reassemble_complexes
from liana.method._pipe_utils._common import _join_stats
from liana.resource._select_resource import _handle_resource
from liana.method._pipe_utils._reassemble_complexes import explode_complexes
from liana.method._pipe_utils._get_mean_perms import _get_means_perms, _get_mat_idx
from liana.method._pipe_utils._aggregate import _aggregate
from liana.method._pipe_utils._common import _get_props
from liana._constants import MethodColumns as M, CommonColumns as C, \
                            PrimaryColumns as P, InternalValues as I

def liana_pipe(adata: anndata.AnnData,
               groupby: str,
               resource_name: str,
               resource: pd.DataFrame | None,
               interactions,
               expr_prop: float,
               min_cells: int,
               base: float,
               de_method: str,
               n_perms: int,
               seed: int,
               verbose: bool,
               use_raw: bool,
               layer: str | None,
               supp_columns: list | None = None,
               return_all_lrs: bool = False,
               _score=None,
               _methods: list = None,
               _consensus_opts: list = None,
               _aggregate_method: str | None = None
               ):
    """
    Parameters
    ----------
    adata
        Annotated data object.
    groupby
        The key of the observations grouping to consider.
    resource_name
        Name of the resource to be loaded and use for ligand-receptor inference.
    resource
        Parameter to enable external resources to be passed. Expects a pandas dataframe
        with [`ligand`, `receptor`] columns. None by default. If provided will overrule
        the resource requested via `resource_name`
    expr_prop
        Minimum expression proportion for the ligands/receptors (and their subunits) in the
         corresponding cell identities. Set to `0`, to return unfiltered results.
    min_cells
        Minimum cells per cell identity
    base
        The base by which to do expm1 (relevant only for 1vsRest logFC calculation)
    de_method
        Differential expression method. `scanpy.tl.rank_genes_groups` is used to rank genes
        according to 1vsRest. The default method is 't-test'. Only relevant if p-values
        are included in `supp_cols`
    n_perms
        n permutations (relevant only for permutation-based methods)
    seed
        Random seed for reproducibility
    verbose
        Verbosity flag
    use_raw
        Use raw attribute of adata if present.
    layer
        Layer in anndata.AnnData.layers to use. If None, use anndata.AnnData.X.
    supp_columns
        additional columns to be added to the output of each method.
    return_all_lrs
        Bool whether to return all LRs, or only those that surpass the expr_prop threshold.
        `False` by default.
    _score
        Instance of Method classes (None by default - returns LR stats - no methods used).
    _methods
        Methods to be run (only relevant for consensus).
    _consensus_opts
        Ways to aggregate interactions across methods by default does all aggregations (['Specificity', 'Magnitude']).
    _aggregate_method
        RobustRankAggregate('rra') or mean rank ('mean').

    Returns
    -------
    A adata frame with ligand-receptor results

    """
    _key_cols = P.primary

    if _score is not None:
        _complex_cols, _add_cols = _score.complex_cols, _score.add_cols
    else:
        _complex_cols = [C.ligand_means, C.receptor_means]
        _add_cols = M.get_all_values()

    if n_perms is None:
        _consensus_opts = 'Magnitude'

    if supp_columns is None:
        supp_columns = []
    _add_cols = _add_cols + [P.ligand, P.receptor,
                             C.ligand_props, C.receptor_props] + supp_columns

    # initialize mat_mean for sca
    mat_mean = None
    mat_max = None

    # Check and Reformat Mat if needed
    assert groupby is not None
    adata = prep_check_adata(adata=adata,
                             groupby=groupby,
                             min_cells=min_cells,
                             use_raw=use_raw,
                             layer=layer,
                             verbose=verbose)

    # get mat mean for SCA (before reducing the features)
    if M.mat_mean in _add_cols:
        mat_mean = np.mean(adata.X, dtype='float32')

    # get mat max for CellChat
    if M.mat_max in _add_cols:
        mat_max = adata.X.max()
        assert isinstance(mat_max, np.float32)

    resource = _handle_resource(interactions=interactions,
                                resource=resource,
                                resource_name=resource_name,
                                verbose=verbose)

    # explode complexes/decomplexify
    resource = explode_complexes(resource)

    # Check overlap between resource and adata
    assert_covered(np.union1d(np.unique(resource[P.ligand]),
                              np.unique(resource[P.receptor])),
                   adata.var_names, verbose=verbose)

    # Filter Resource
    resource = filter_resource(resource, adata.var_names)

    if verbose:
        print(f"Generating ligand-receptor stats for {adata.shape[0]} samples "
              f"and {adata.shape[1]} features")

    # Create Entities
    entities = np.union1d(np.unique(resource[P.ligand]),
                          np.unique(resource[P.receptor]))

    # Filter to only include the relevant genes
    adata = adata[:, np.intersect1d(entities, adata.var.index)]

    # Get lr results
    lr_res = _get_lr(adata=adata,
                     resource=resource,
                     mat_mean=mat_mean,
                     mat_max=mat_max,
                     relevant_cols=_key_cols + _add_cols + _complex_cols,
                     de_method=de_method,
                     base=base,
                     verbose=verbose
                     )

    # Mean Sums required for NATMI (note done on subunits also)
    if M.ligand_means_sums in _add_cols:
        on = [x for x in P.complete if x != P.source]
        lr_res = _sum_means(lr_res, what=C.ligand_means, on=on)
    if M.receptor_means_sums in _add_cols:
        on = [x for x in P.complete if x != P.target]
        lr_res = _sum_means(lr_res, what=C.receptor_means, on=on)

    # Calculate Score
    if _score is not None:
        if _score.method_name == "Rank_Aggregate":
            # Run all methods in consensus
            lrs = {}
            for method in _methods:
                if verbose:
                    print(f"Running {method.method_name}")

                lrs[method.method_name] = \
                    _run_method(lr_res=lr_res.copy(),
                                adata=adata,
                                expr_prop=expr_prop,
                                _score=method,
                                _key_cols=_key_cols,
                                _complex_cols=method.complex_cols,
                                _add_cols=method.add_cols,
                                n_perms=n_perms,
                                seed=seed,
                                return_all_lrs=return_all_lrs,
                                verbose=verbose,
                                _aggregate_flag=True
                                )
            if _consensus_opts is not False:
                lr_res = _aggregate(lrs,
                                    consensus=_score,
                                    aggregate_method=_aggregate_method,
                                    _key_cols=_key_cols,
                                    _consensus_opts=_consensus_opts,
                                    )
            else:  # Return by method results as they are
                return lrs
        else:  # Run the specific method in mind
            lr_res = _run_method(lr_res=lr_res,
                                 adata=adata,
                                 expr_prop=expr_prop,
                                 _score=_score, _key_cols=_key_cols,
                                 _complex_cols=_complex_cols,
                                 _add_cols=_add_cols,
                                 n_perms=n_perms,
                                 return_all_lrs=return_all_lrs,
                                 verbose=verbose,
                                 seed=seed)
    else:  # Just return lr_res
        lr_res = filter_reassemble_complexes(lr_res=lr_res,
                                             _key_cols=_key_cols,
                                             expr_prop=expr_prop,
                                             complex_cols=_complex_cols,
                                             return_all_lrs=return_all_lrs)

    if _score is not None:
        orderby, ascending =  (_score.magnitude, _score.magnitude_ascending) if _score.magnitude is not None \
            else (_score.specificity, _score.specificity_ascending)

        lr_res = lr_res.sort_values(by=orderby, ascending=ascending)

    return lr_res


def _get_lr(adata, resource, relevant_cols, mat_mean, mat_max, de_method, base, verbose):
    labels = adata.obs[I.label].cat.categories

    # Method-specific stats
    connectome_flag = (M.ligand_zscores in relevant_cols) | (
                M.receptor_zscores in relevant_cols)
    if connectome_flag:
        adata.layers['scaled'] = sc.pp.scale(adata, copy=True).X

    logfc_flag = (M.ligand_logfc in relevant_cols) | (
            M.receptor_logfc in relevant_cols)
    if logfc_flag:
        if 'log1p' in adata.uns_keys():
            if (adata.uns['log1p']['base'] is not None) & verbose:
                print("Assuming that counts were `natural` log-normalized!")
        elif ('log1p' not in adata.uns_keys()) & verbose:
            print("Assuming that counts were `natural` log-normalized!")
        adata.layers['normcounts'] = adata.X.copy()
        adata.layers['normcounts'].data = _expm1_base(adata.X.data, base)

    # initialize dict
    dedict = {}

    # Calc pvals + other stats per gene or not
    rank_genes_bool = (C.ligand_pvals in relevant_cols) | (C.receptor_pvals in relevant_cols)
    if rank_genes_bool:
        adata = sc.tl.rank_genes_groups(adata, groupby=I.label,
                                        method=de_method, use_raw=False,
                                        copy=True)

    for label in labels:
        temp = adata[adata.obs[I.label] == label, :]
        a = _get_props(temp.X)
        stats = pd.DataFrame({'names': temp.var_names, 'props': a}). \
            assign(label=label).sort_values('names')
        if rank_genes_bool:
            pvals = sc.get.rank_genes_groups_df(adata, label)
            stats = stats.merge(pvals)
        dedict[label] = stats

    # check if genes are ordered correctly
    if not list(adata.var_names) == list(dedict[labels[0]]['names']):
        raise AssertionError("Variable names did not match DE results!")

    # Calculate Mean, logFC and z-scores by group
    for label in labels:
        temp = adata[adata.obs[I.label].isin([label])]
        dedict[label]['means'] = temp.X.mean(axis=0).A.flatten()
        if connectome_flag:
            dedict[label]['zscores'] = temp.layers['scaled'].mean(axis=0)
        if logfc_flag:
            dedict[label]['logfc'] = _calc_log2fc(adata, label)
        if isinstance(mat_max, np.float32):  # cellchat flag
            dedict[label]['trimean'] = _trimean(temp.X / mat_max)

    # Create df /w cell identity pairs
    # pd.DataFrame(list(product(labels, labels))); TODO: check this
    pairs = (pd.DataFrame(np.array(np.meshgrid(labels, labels))
                          .reshape(2, np.size(labels) * np.size(labels)).T)
             .rename(columns={0: P.source, 1: P.target}))

    # Join Stats
    lr_res = pd.concat(
        [_join_stats(source, target, dedict, resource) for source, target in
         zip(pairs[P.source], pairs[P.target])]
    )

    if M.mat_mean in relevant_cols:
        assert isinstance(mat_mean, np.float32)
        lr_res[M.mat_mean] = mat_mean

    # NOTE: this is not needed
    if isinstance(mat_max, np.float32):
        lr_res[M.mat_max] = mat_max

    # subset to only relevant columns
    relevant_cols = np.intersect1d(relevant_cols, lr_res.columns)

    return lr_res[relevant_cols]


def _sum_means(lr_res, what, on):
    return lr_res.join(lr_res.groupby(on)[what].sum(), on=on, rsuffix='_sums')


def _calc_log2fc(adata, label) -> np.ndarray:
    # Get subject vs rest cells
    subject = adata[adata.obs[I.label].isin([label])]
    rest = adata[~adata.obs[I.label].isin([label])]

    # subject and rest means
    subj_means = subject.layers['normcounts'].mean(0).A.flatten()
    rest_means = rest.layers['normcounts'].mean(0).A.flatten()

    # log2 + 1 transform
    subj_log2means = np.log2(subj_means + 1)
    loso_log2means = np.log2(rest_means + 1)

    logfc_vec = subj_log2means - loso_log2means

    return logfc_vec


def _expm1_base(X, base):
    return np.power(base, X) - 1


def _run_method(lr_res: pandas.DataFrame,
                adata: anndata.AnnData,
                expr_prop: float,
                _score,
                _key_cols: list,
                _complex_cols: list,
                _add_cols: list,
                n_perms: int,
                seed: int,
                return_all_lrs: bool,
                verbose: bool,
                _aggregate_flag: bool = False  # Indicates whether we're generating the consensus
                ) -> pd.DataFrame:
    # re-assemble complexes - specific for each method
    lr_res = filter_reassemble_complexes(lr_res=lr_res,
                                         _key_cols=_key_cols,
                                         expr_prop=expr_prop,
                                         return_all_lrs=return_all_lrs,
                                         complex_cols=_complex_cols)

    _add_cols = _add_cols + [P.ligand, P.receptor]
    relevant_cols = reduce(np.union1d, [_key_cols, _complex_cols, _add_cols])
    if return_all_lrs:
        relevant_cols = list(relevant_cols) + [I.lrs_to_keep]
        # separate those that pass from rest
        rest_res = lr_res[~lr_res[I.lrs_to_keep]]
        rest_res = rest_res[relevant_cols]
        lr_res = lr_res[lr_res[I.lrs_to_keep]]
    lr_res = lr_res[relevant_cols]

    if (M.mat_max in _add_cols) & (_score.method_name == "CellChat"):
        # CellChat matrix_max
        norm_factor = np.unique(lr_res[M.mat_max].values)[0]
        agg_fun = _trimean
    else:
        norm_factor = None
        agg_fun = np.mean

    if _score.permute:
        # get permutations
        if n_perms is not None:
            perms = _get_means_perms(adata=adata,
                                     n_perms=n_perms,
                                     seed=seed,
                                     agg_fun=agg_fun,
                                     norm_factor=norm_factor,
                                     verbose=verbose)
            # get tensor indexes for ligand, receptor, source, target
            ligand_idx, receptor_idx, source_idx, target_idx = _get_mat_idx(adata, lr_res)

            # ligand and receptor perms
            ligand_stat_perms = perms[:, source_idx, ligand_idx]
            receptor_stat_perms = perms[:, target_idx, receptor_idx]
            # stack them together
            perm_stats = np.stack((ligand_stat_perms, receptor_stat_perms), axis=0)
        else:
            perm_stats = None
            _score.specificity = None

        scores = _score.fun(x=lr_res,
                            perm_stats=perm_stats)
    else:  # non-perm funs
        scores = _score.fun(x=lr_res)

    lr_res.loc[:, _score.magnitude] = scores[0]
    lr_res.loc[:, _score.specificity] = scores[1]


    if return_all_lrs:
        # re-append rest of results
        lr_res = pd.concat([lr_res, rest_res], copy=False)
        if _score.magnitude is not None:
            fill_value = _assign_min_or_max(lr_res[_score.magnitude],
                                            _score.magnitude_ascending)
            lr_res.loc[~lr_res[I.lrs_to_keep], _score.magnitude] = fill_value
        if _score.specificity is not None:
            fill_value = _assign_min_or_max(lr_res[_score.specificity],
                                            _score.specificity_ascending)
            lr_res.loc[~lr_res[I.lrs_to_keep], _score.specificity] = fill_value

    if _aggregate_flag:  # if consensus keep only the keys and the method scores
        lr_res = lr_res[_key_cols + [_score.magnitude, _score.specificity]]

    # remove redundant cols for some scores
    if (_score.magnitude is None) | (_score.specificity is None):
        lr_res = lr_res.drop([None], axis=1)

    return lr_res


def _assign_min_or_max(x, x_ascending):
    if x_ascending:
        return np.max(x)
    else:
        return np.min(x)


def _trimean(a, axis=0):
    return np.mean(
        np.quantile(a.A,
                    q=[0.25, 0.5, 0.5, 0.75],
                    axis=axis),
        axis=axis)



from __future__ import annotations

import numpy as np

from anndata import AnnData
from pandas import DataFrame

from ._common import _process_scores
from liana._logging import _check_if_installed
from liana._docs import d
from liana._constants import DefaultValues as V, Keys as K, PrimaryColumns as P

@d.dedent
def to_tensor_c2c(adata:AnnData=None,
                  sample_key:str = None,
                  score_key:str = None,
                  liana_res: (DataFrame or None) = None,
                  source_key:str = P.source,
                  target_key:str = P.target,
                  ligand_key:str = P.ligand_complex,
                  receptor_key:str = P.receptor_complex,
                  uns_key:str = K.uns_key,
                  non_expressed_fill:(float or None) = None,
                  inverse_fun: callable = V.inverse_fun,
                  non_negative:bool = True,
                  return_dict:bool = False,
                  **kwargs
                  ):
    """
    Function to convert a LIANA result to a tensor for cell2cell analysis.

    Parameters
    ----------

    %(adata)s
    %(sample_key)s
    %(score_key)s
    liana_res
        A dataframe with the LIANA results. If None, it will be taken from `adata.uns[uns_key]`.
    %(source_key)s
    %(target_key)s
    %(ligand_key)s
    %(receptor_key)s
    %(uns_key)s
    %(inverse_fun)s
    non_expressed_fill : `float`, optional (default: None)
        Value to fill for non-expressed ligand-receptor pairs.
    non_negative : `bool`, optional (default: True)
        Whether to make the tensor non-negative.
    return_dict : `bool`, optional (default: False)
        Whether to return a dictionary of tensors.
    **kwargs : keyword arguments to pass to Tensor-cell2cell's `cell2cell.tensor.external_scores.dataframes_to_tensor` function.

    Returns
    -------
    Returns a tensor of shape (n_samples, n_senders, n_receivers, n_interactions) or a dictionary of tensors if `return_dict` is True.


    """

    # check if cell2cell is installed
    c2c = _check_if_installed("cell2cell")

    if (liana_res is None) & (adata is None):
        raise AttributeError('Ambiguous! One of `liana_res` or `adata` should be provided.')
    if adata is not None:
        assert uns_key in adata.uns_keys()
        liana_res = adata.uns[uns_key].copy()
    if liana_res is not None:
        liana_res = liana_res.copy()
    if (liana_res is None) & (adata is None):
        raise ValueError('`liana_res` or `adata` must be provided!')

    keys = np.array([sample_key, source_key, target_key, ligand_key, receptor_key])
    missing_keys = keys[[ key not in liana_res.columns for key in keys]]

    if any(missing_keys):
        raise ValueError(f'`{missing_keys}` not found in `adata.uns[{uns_key}]`! Please check your input.')

    # remove unneeded columns
    keys = [sample_key, source_key, target_key, ligand_key, receptor_key, score_key]
    keys = keys + ['lrs_to_keep'] if 'lrs_to_keep' in liana_res.columns else keys
    liana_res = liana_res[keys]

    # check for duplicates
    if liana_res[[sample_key, source_key, target_key, ligand_key, receptor_key]].duplicated().any():
        raise ValueError("Duplicate rows found in the input data")

    liana_res = _process_scores(liana_res, score_key, inverse_fun)

    # set negative to 0
    if non_negative:
        liana_res[score_key] = liana_res[score_key].clip(lower=0)

    # set non-expressed to 0 (if lrs_to_keep column is present)
    if ('lrs_to_keep' in liana_res.columns) & (non_expressed_fill is not None):
        liana_res.loc[~liana_res['lrs_to_keep'], score_key] = non_expressed_fill

    # split into dictionary by sample
    liana_res = {sample:df for sample, df in liana_res.groupby(sample_key)}

    if return_dict:
        return liana_res

    tensor = c2c.tensor.dataframes_to_tensor(liana_res,
                                             sender_col=source_key,
                                             receiver_col=target_key,
                                             ligand_col=ligand_key,
                                             receptor_col=receptor_key,
                                             score_col=score_key,
                                             **kwargs)

    return tensor



from liana.testing._sample_anndata import generate_toy_mdata
from liana.utils import mdata_to_anndata, zi_minmax, neg_to_zero
from numpy.testing import assert_almost_equal

mdata = generate_toy_mdata()

def test_m_to_adata():
    adata = mdata_to_anndata(mdata, x_mod='adata_x', y_mod='adata_y',
                            x_transform=False, y_transform=False, verbose=True)
    assert adata.shape == mdata.shape
    assert_almost_equal(adata.X.max(), 3.431)


def test_mdata_transformations():
    # test minmax
    adata = mdata_to_anndata(mdata, x_mod='adata_x', y_mod='adata_y',
                             x_transform=zi_minmax, y_transform=zi_minmax,
                             verbose=False)
    assert adata.X.max() == 1
    assert_almost_equal(adata.X.sum(), 2120.704, decimal=3)

    # test cutoff
    def zi_minmax_cutoff(x):
        x = zi_minmax(x, cutoff=0.5)
        return x

    adata = mdata_to_anndata(mdata, x_mod='adata_x', y_mod='adata_y',
                             x_transform=zi_minmax_cutoff, y_transform=zi_minmax_cutoff,
                             verbose=False)
    assert_almost_equal(adata.X.sum(), 1497.3386, decimal=4)

    # test non-negative
    from scanpy.pp import scale
    scale(mdata.mod['adata_x'])

    adata = mdata_to_anndata(mdata, x_mod='adata_x', y_mod='adata_y',
                             x_transform=neg_to_zero, y_transform=False,
                             verbose=False)
    assert_almost_equal(adata.X.max(), 7.760507, decimal=5)
    assert_almost_equal(adata.X.min(), 0, decimal=5)



import numpy as np
from itertools import product

from liana.testing._sample_anndata import generate_toy_mdata
from liana.method.sp._SpatialBivariate import bivar

mdata = generate_toy_mdata()
interactions = list(product(mdata.mod['adata_x'].var.index,
                            mdata.mod['adata_y'].var.index))
ones = np.ones((mdata.shape[0], mdata.shape[0]), dtype=np.float64)
mdata.obsp['ones'] = ones


def test_bivar_morans():
    bivar(mdata,
          x_mod='adata_x',
          y_mod='adata_y',
          function_name='morans',
          x_use_raw=False,
          y_use_raw=False,
          interactions=interactions
          )
    assert 'local_scores' in mdata.mod.keys()
    np.testing.assert_almost_equal(mdata.mod['local_scores'].X.sum(), -45.86797, decimal=5)


def test_bivar_morans_perms():
    bivar(mdata,
          x_mod='adata_x',
          y_mod='adata_y',
          function_name='morans',
          n_perms=2,
          x_use_raw=False,
          y_use_raw=False,
          interactions=interactions)


    assert 'local_scores' in mdata.mod.keys()
    local_pvals = mdata.mod['local_scores'].layers['pvals']
    np.testing.assert_almost_equal(np.mean(local_pvals), 0.6112173202614387, decimal=6)


def test_bivar_nondefault():
    global_stats, local_scores = \
          bivar(mdata,
                x_mod='adata_x',
                y_mod='adata_y',
                function_name='morans',
                n_perms=0,
                connectivity_key='ones',
                remove_self_interactions=False,
                x_layer = "scaled",
                y_layer = "scaled",
                x_use_raw=False,
                y_use_raw=False,
                inplace=False,
                add_categories=True,
                interactions=interactions
                )

    # if all are the same weights, then everything is close to 0?
    np.testing.assert_almost_equal(global_stats['global_r'].sum(), 0)
    local_scores.shape == (700, 100)
    np.testing.assert_almost_equal(np.min(np.min(local_scores.layers['pvals'])), 0.5, decimal=2)

    assert local_scores.layers['cats'].sum() == -8160



def test_bivar_adata():
      mdata.mod['adata_x'].obsp = mdata.obsp

      bivar(mdata=mdata.mod['adata_x'],
            x_mod=None,
            y_mod=None,
            x_use_raw=False,
            y_use_raw=False,
            function_name='morans',
            connectivity_key='ones',
            interactions=interactions)


def test_masked_spearman():
    bivar(mdata,
          x_mod='adata_x',
          y_mod='adata_y',
          x_use_raw=False,
          y_use_raw=False,
          function_name='masked_spearman',
          interactions=interactions,
          connectivity_key='ones')

    # check local
    assert 'local_scores' in mdata.mod.keys()
    np.testing.assert_almost_equal(mdata.mod['local_scores'].X.mean(), 0.18438724, decimal=5)

    # check global
    assert 'global_res' in mdata.uns.keys()
    global_res = mdata.uns['global_res']
    assert set(['global_mean','global_sd']).issubset(global_res.columns)
    np.testing.assert_almost_equal(global_res['global_mean'].mean(), 0.18438746, decimal=5)
    np.testing.assert_almost_equal(global_res['global_sd'].mean(), 8.498836e-07, decimal=5)


def test_vectorized_pearson():
    bivar(mdata,
          x_mod='adata_x',
          y_mod='adata_y',
          x_use_raw=False,
          y_use_raw=False,
          function_name='pearson',
          n_perms=100,
          interactions=interactions)

    # check local
    assert 'local_scores' in mdata.mod.keys()
    adata = mdata.mod['local_scores']
    np.testing.assert_almost_equal(adata.X.mean(), 0.0011550441, decimal=5)
    np.testing.assert_almost_equal(adata.layers['pvals'].mean(), 0.755160947712419, decimal=3)

    # check global
    assert 'global_res' in mdata.uns.keys()
    global_res = mdata.uns['global_res']
    assert set(['global_mean','global_sd']).issubset(global_res.columns)
    np.testing.assert_almost_equal(global_res['global_mean'].mean(), 0.0011550438183169959, decimal=5)
    np.testing.assert_almost_equal(global_res['global_sd'].mean(), 0.3227660823917939, decimal=5)



import pandas as pd
import numpy as np
from liana._logging import _logg, _check_if_installed
from liana._constants import DefaultValues as V


def build_prior_network(ppis,
                        input_nodes,
                        output_nodes,
                        lr_sep=None,
                        verbose=V.verbose):
    """
    Build Prior Network from PPIs and input/output nodes.

    Parameters
    ----------
    ppis : list of tuples or pandas DataFrame
        The PPIs to use for the prior network. If a pandas DataFrame is provided, it must have the columns
    input_nodes : dict
        A dictionary of input nodes. The keys are the node names, the values are the node scores.
    output_nodes : dict
        A dictionary of output nodes. The keys are the node names, the values are the node scores.
    lr_sep : str, optional
        The separator to use to split the input nodes into ligand and receptor. If None, the input nodes will be used as is.
    verbose : bool, optional
        Whether to print progress information. Default: True

    Returns
    -------
    corneto.Graph

    """

    cn = _check_if_installed("corneto")

    if lr_sep is not None:
        if any(lr_sep in k for k in input_nodes.keys()):
            _logg(f"Input nodes are in the format Ligand{lr_sep}Receptor. Extracting only the receptor...", verbose=verbose)
            # Print only at most the first 3 entries
            for k, v in list(input_nodes.items())[:3]:
                _logg(f" - {k} -> {v}", verbose=verbose)
            _logg(" - ...", verbose=verbose)
            # Do the split only if lr_sep is present, otherwise get the same key:
            input_nodes = {k.split(lr_sep)[-1]: v for k, v in input_nodes.items()}

    _logg("Importing network...", verbose=verbose)
    if isinstance(ppis, list):
        G = cn.Graph.from_sif_tuples(ppis)
    elif isinstance(ppis, pd.DataFrame):
        ppis = [(r.source, r.mor, r.target) for (_, r) in ppis.iterrows() if r.source != r.target]
        G = cn.Graph.from_sif_tuples(ppis)
    else:
        raise ValueError("PPIs must be a list of tuples or a pandas DataFrame.")
    _logg("done.", verbose=verbose)

    incl_inputs = set(G.vertices).intersection(set(input_nodes))
    incl_outputs = set(G.vertices).intersection(set(output_nodes))

    _logg(f" - Nodes x Edges: {G.num_vertices, G.num_edges}", verbose=verbose)
    _logg(f" - Provided inputs included in the prior network: {len(incl_inputs)}/{len(input_nodes)}", verbose=verbose)
    _logg(f" - Provided outputs included in the network: {len(incl_outputs)}/{len(output_nodes)}", verbose=verbose)

    _logg("Performing reachability analysis...", verbose=verbose)
    Gp = G.prune(list(incl_inputs), list(incl_outputs))
    _logg("done.", verbose=verbose)
    pruned_size = (Gp.num_vertices, Gp.num_edges)
    incl_inputs_pruned = set(Gp.vertices).intersection(incl_inputs)
    incl_outputs_pruned = set(Gp.vertices).intersection(incl_outputs)

    _logg(f" - Selected inputs: {len(incl_inputs_pruned)}/{len(incl_inputs)}.", verbose=verbose)
    _logg(f" - Selected outputs: {len(incl_outputs_pruned)}/{len(incl_outputs)}.", verbose=verbose)
    _logg(f" - Final size of the prior graph: {pruned_size}.", verbose=verbose)

    if len(incl_outputs_pruned) == 0:
        raise ValueError("None of the output nodes can be reached from the provided input nodes in the PPI network.")
    if len(incl_inputs_pruned) == 0:
        raise ValueError("None of the input nodes can reach any of the output nodes in the PPI network.")

    return Gp

def _get_scores(d):
    return (
       [v for v in d.values() if v < 0],
       [v for v in d.values() if v > 0]
    )


def find_causalnet(
        prior_graph,
        input_node_scores,
        output_node_scores,
        node_weights=None,
        node_cutoff=0.1,
        min_penalty=0.01,
        max_penalty=1.0,
        missing_penalty=10,
        edge_penalty=0.01,
        solver=None,
        seed=1337,
        max_seconds=None,
        verbose=True,
        **kwargs
        ):
    """
    Find the causal network that best explains the input/output node scores.

    Parameters
    ----------
    prior_graph : corneto.Graph
        The prior graph to use for the search.
    input_node_scores : dict
        A dictionary of input node scores.
    output_node_scores : dict
        A dictionary of output node scores.
    node_weights : dict, optional
        A dictionary of node weights. The keys are the node names, the values are the weeights.
        If None, all nodes will have the same weight.
    node_cutoff : float
        The cutoff to use for the node weights. Nodes with a weight below this cutoff will be assigned
        the max_penalty, nodes with a weight above this cutoff will be assigned the min_penalty.
        Only used if node_weights is not None. Default: 0.1
    min_penalty : float
        The minimum penalty to assign to nodes with a weight above the cutoff.
        Only used if node_weights is not None. Default: 0.01
    max_penalty : float
        The maximum penalty to assign to nodes with a weight below the cutoff
        Only used if node_weights is not None. Default: 1.0
    missing_penalty : float
        The penalty to assign to nodes that are not measured. Default: 10
    edge_penalty : float
        The penalty to assign to edges. Default: 0.01
    solver : str, optional
        The solver to use. If None, the default solver will be used. Default: None
        It will default to the solver included in SCIPY, if no other solver is available.
    seed : int, optional
        The seed to use for the random number generator. Default: 1337
    max_seconds : int, optional
        The maximum number of seconds to run the solver. Default: None
    verbose : bool, optional
        Whether to print progress information. Default: True
    **kwargs : dict, optional
        Additional arguments to pass to the solver.
    """

    cn = _check_if_installed("corneto")

    if solver is None:
        solver = cn.methods.carnival.select_mip_solver()

    measured_nodes = set(input_node_scores.keys()) | set(output_node_scores.keys())

    _logg("Total positive/negative scores of the inputs and outputs:", verbose=verbose)
    w_neg_in, w_pos_in = _get_scores(input_node_scores)
    w_neg_out, w_pos_out = _get_scores(output_node_scores)
    _logg(f" - (-) input nodes: {sum(w_neg_in)}", verbose=verbose)
    _logg(f" - (+) input nodes: {sum(w_pos_in)}", verbose=verbose)
    _logg(f" - (-) output nodes: {sum(w_neg_out)}", verbose=verbose)
    _logg(f" - (+) output nodes: {sum(w_pos_out)}", verbose=verbose)

    # Total weights
    total = abs(sum(w_neg_in)) + abs(sum(w_neg_out)) + sum(w_pos_in) + sum(w_pos_out)
    _logg(f" - abs total (inputs + outputs): {total}", verbose=verbose)

    if node_weights is None:
        node_penalties = {}
    else:
        node_penalties = _weights_to_penalties(node_weights,
                                               cutoff=node_cutoff,
                                               max_penalty=max_penalty,
                                               min_penalty=min_penalty)

    # assign 0 penalties to input/output nodes, missing_penalty to missing nodes
    # add a small amount of noise to the penalties to ensure reproducible solutions
    rng = np.random.default_rng(seed=seed)
    c_node_penalties = {k: node_penalties.get(k, missing_penalty) + rng.uniform(0, 0.00001)
                        if k not in measured_nodes else 0.0 for k in prior_graph.vertices}

    _logg("Building CORNETO problem...", verbose=verbose)
    P, G = cn.methods.carnival._extended_carnival_problem(
        prior_graph,
        input_node_scores,
        output_node_scores,
        node_penalties=c_node_penalties,
        edge_penalty=edge_penalty
    )

    _logg(f"Solving with {solver}...", verbose=verbose)
    if (solver=='scipy') and verbose:
        kwargs.update(scipy_options=dict(disp='true'))
    P.solve(
        solver=solver,
        max_seconds=max_seconds,
        verbosity=int(verbose),
        **kwargs)

    _logg("Done.", verbose=verbose)

    obj_names = ["Loss (unfitted inputs/output)", "Edge penalty error", "Node penalty error"]
    _logg("Solution summary:", verbose=verbose)
    for s, o in zip(obj_names, P.objectives):
        _logg(f" - {s}: {o.value}", verbose=verbose)
    rows, cols = cn.methods.carnival.export_results(P, G, input_node_scores, output_node_scores)
    df = pd.DataFrame(rows, columns=cols)
    return df, P


def _weights_to_penalties(props,
                          cutoff,
                          min_penalty,
                          max_penalty):
    if any(p < 0 or p > 1 for p in props.values()):
        raise ValueError("Node weights were not between 0 and 1. Consider minmax or another normalization.")

    return {k: max_penalty if v < cutoff else min_penalty for k, v in props.items()}



def _join_stats(source, target, dedict, resource):
    """
    Joins and renames source-ligand and target-receptor stats to the ligand-receptor resource

    Parameters
    ----------
    source
        Source/Sender cell type
    target
        Target/Receiver cell type
    dedict
        dictionary
    resource
        Ligand-receptor Resource

    Returns
    -------
    Ligand-Receptor stats

    """
    source_stats = dedict[source].copy()
    source_stats.columns = source_stats.columns.map(
        lambda x: 'ligand_' + str(x))
    source_stats = source_stats.rename(
        columns={'ligand_names': 'ligand', 'ligand_label': 'source'})

    target_stats = dedict[target].copy()
    target_stats.columns = target_stats.columns.map(
        lambda x: 'receptor_' + str(x))
    target_stats = target_stats.rename(
        columns={'receptor_names': 'receptor', 'receptor_label': 'target'})

    bound = resource.merge(source_stats).merge(target_stats)

    return bound

def _get_props(X_mask):
    return X_mask.getnnz(axis=0) / X_mask.shape[0]



import numpy as np
from liana.method.sc._Method import Method, MethodMeta
from liana.method._pipe_utils._get_mean_perms import _calculate_pvals


# simplified/resource-generalizable cellchat probability score
def _lr_probability(perm_stats, axis=0):
    lr_prob = np.product(perm_stats, axis=axis)

    return lr_prob / (0.5 + lr_prob)  # Kh=0.5


# Internal Function to calculate CellPhoneDB LR_mean and p-values
def _cellchat_score(x, perm_stats) -> tuple:
    """
    Calculate CellChat-like LR means and p-values

    Parameters
    ----------
    x
        DataFrame with LIANA results
    perm_stats
        Permutation statistics (2 (ligand-receptor), n_perms (number of permutations, n_rows in lr_res)

    Returns
    -------
    A tuple with lr_mean and pvalue for x

    """
    lr_prob = _lr_probability((x['ligand_trimean'].values, x['receptor_trimean'].values))
    cellchat_pvals = _calculate_pvals(lr_prob, perm_stats, _lr_probability)

    return lr_prob, cellchat_pvals


# Initialize CellChat Meta
_cellchat = MethodMeta(method_name="CellChat",
                       complex_cols=["ligand_trimean", "receptor_trimean"],
                       add_cols=['mat_max'],
                       fun=_cellchat_score,
                       magnitude="lr_probs",
                       magnitude_ascending=False,
                       specificity="cellchat_pvals",
                       specificity_ascending=True,
                       permute=True,
                       reference="Jin, S., Guerrero-Juarez, C.F., Zhang, L., Chang, I., Ramos, "
                                 "R., Kuan, C.H., Myung, P., Plikus, M.V. and Nie, Q., "
                                 "2021. Inference and analysis of cell-cell communication using "
                                 "CellChat. Nature communications, 12(1), pp.1-20. "
                       )

cellchat = Method(_method=_cellchat)



from __future__ import annotations

import numpy as np
from pandas import DataFrame, Series

from scipy.sparse import csr_matrix, isspmatrix_csr, hstack
from scipy.stats import norm
from tqdm import tqdm
from anndata import AnnData


def _rename_means(lr_stats, entity):
    df = lr_stats.copy()
    df.columns = df.columns.map(lambda x: entity + '_' + str(x) if x != 'gene' else 'gene')
    return df.rename(columns={'gene': entity})


def _local_to_dataframe(array, idx, columns):
    if array is None:
        return None
    return DataFrame(array.T, index=idx, columns=columns)


def _local_permutation_pvals(x_mat, y_mat, weight, local_truth, local_fun, n_perms, seed,
                             mask_negatives, local_msk, verbose):
    """
    Calculate local pvalues for a given local score function.

    Parameters
    ----------
    x_mat
        2D array with x variables
    y_mat
        2D array with y variables
    local_truth
        2D array with non-permuted local scores/co-expressions
    weight
        connectivity weights
    n_perms
        number of permutations
    seed
        Reproducibility seed
    mask_negatives
        Whether to mask negative correlations pvalue

    Returns
    -------
    2D array with shape(n_spot, xy_n)

    """
    rng = np.random.default_rng(seed)

    xy_n = local_truth.shape[0]
    spot_n = local_truth.shape[1]

    # permutation cubes to be populated
    local_pvals = np.zeros((xy_n, spot_n))

    # shuffle the matrix
    for i in tqdm(range(n_perms), disable=not verbose):
        _idx = rng.permutation(spot_n)
        perm_score = local_fun(x_mat=x_mat[_idx, :], y_mat=y_mat, weight=weight)
        if mask_negatives:
            local_pvals += np.array(perm_score >= local_truth, dtype=int)
        else:
            local_pvals += (np.array(np.abs(perm_score) >= np.abs(local_truth), dtype=int))

    local_pvals = local_pvals / n_perms

    if mask_negatives:
        local_pvals[~local_msk] = 1

    return local_pvals


def _zscore(mat, local=True, axis=0):
    spot_n = mat.shape[1]

    # NOTE: specific to global SpatialDM
    if not local:
        spot_n = 1

    mat = np.array(mat - np.array(mat.mean(axis=axis)))
    mat = mat / np.sqrt(np.sum(mat ** 2, axis=axis, keepdims=True) / spot_n)

    return mat


def _encode_cats(a, weight):
    # if only positive
    if np.all(a >= 0):
        a = _zscore(a.T).T
    a = a @ weight
    a = np.where(a > 0, 1, np.where(a < 0, -1, np.nan))
    return a

def _categorize(x_mat, y_mat, weight, idx, columns):
    x_cats = _encode_cats(x_mat.A, weight)
    y_cats = _encode_cats(y_mat.A, weight)

    # add the two categories, and simplify them to ints
    cats = x_cats + y_cats
    cats = np.where(cats == 2, 1, np.where(cats == 0, -1, 0))

    return cats


def _global_permutation_pvals(x_mat, y_mat, weight, global_r, n_perms, mask_negatives, seed, verbose):
    """
    Calculate permutation pvals

    Parameters
    ----------
    x_mat
        Matrix with x variables
    y_mat
        Matrix with y variables
    dist
        connectivity weights 2D array
    global_r
        Global Moran's I, 1D array
    n_perms
        Number of permutations
    mask_negatives
        Whether to mask negative p-values
    seed
        Reproducibility seed

    Returns
    -------
    1D array with same size as global_r

    """
    assert isinstance(weight, csr_matrix)
    rng = np.random.default_rng(seed)

    # initialize mat /w n_perms * number of X->Y
    idx = x_mat.shape[1]

    # permutation mat /w n_permss x LR_n
    perm_mat = np.zeros((n_perms, global_r.shape[0]))

    for perm in tqdm(range(n_perms), disable=not verbose):
        _idx = rng.permutation(idx)
        perm_mat[perm, :] = ((x_mat[:, _idx] @ weight) * y_mat).sum(axis=1)  # flipped x_mat

    if mask_negatives:
        global_pvals = 1 - (global_r > perm_mat).sum(axis=0) / n_perms
    else:
        # TODO Proof this makes sense
        global_pvals = 2 * (1 - (np.abs(global_r) > np.abs(perm_mat)).sum(axis=0) / n_perms)

    return global_pvals


def _global_zscore_pvals(weight, global_r, mask_negatives):
    """

    Parameters
    ----------
    weight
        connectivity weight matrix (spot_n x spot_n)
    global_r
        Array with
    mask_negatives: bool
        whether to mask negative correlation p-values

    Returns
    -------
        1D array with the size of global_r

    """
    weight = np.array(weight.todense())
    spot_n = weight.shape[0]

    # global distance/weight variance as in spatialDM
    numerator = spot_n ** 2 * ((weight * weight).sum()) - \
                (2 * spot_n * (weight @ weight).sum()) + \
                (weight.sum() ** 2)
    denominator = spot_n ** 2 * (spot_n - 1) ** 2
    weight_var_sq = (numerator / denominator) ** (1 / 2)

    global_zscores = global_r / weight_var_sq

    if mask_negatives:
        global_zpvals = norm.sf(global_zscores)
    else:
        global_zpvals = norm.sf(np.abs(global_zscores)) * 2

    return global_zpvals


def _local_zscore_pvals(x_mat, y_mat, local_truth, weight, mask_negatives, local_msk):
    """

    Parameters
    ----------
    x_mat
        2D array with x variables
    y_mat
        2D array with y variables
    local_r
        2D array with Local Moran's I
    weight
        connectivity weights
    mask_negatives
        Whether to mask negative correlations pvalue

    Returns
    -------
    2D array of p-values with shape(n_spot, xy_n)

    """
    spot_n = weight.shape[0]

    x_norm = np.apply_along_axis(norm.fit, axis=0, arr=x_mat)
    y_norm = np.apply_along_axis(norm.fit, axis=0, arr=y_mat)

    # get x,y std
    x_sigma, y_sigma = x_norm[1, :], y_norm[1, :]

    x_sigma = x_sigma * spot_n / (spot_n - 1)
    y_sigma = y_sigma * spot_n / (spot_n - 1)

    std = _get_local_var(x_sigma, y_sigma, weight, spot_n)
    local_zscores = local_truth / std

    if mask_negatives:
        local_zpvals = norm.sf(local_zscores)
        local_zpvals[~local_msk] = 1
    else:
        local_zpvals = norm.sf(np.abs(local_zscores))

    return local_zpvals


def _get_local_var(x_sigma, y_sigma, weight, spot_n):
    """
    Spatial weight variance as in spatialDM (Li et al., 2022)

    Parameters
    ----------
    x_sigma
        Standard deviations for each x (e.g. std of all ligands in the matrix)
    y_sigma
        Standard deviations for each y (e.g. std of all receptors in the matrix)
    weight
        connectivity weight matrix
    spot_n
        number of spots/cells in the matrix

    Returns
    -------
    2D array of standard deviations with shape(n_spot, xy_n)

    """
    weight_sq = (np.array(weight.todense()) ** 2).sum(axis=1)

    n_weight = 2 * (spot_n - 1) ** 2 / spot_n ** 2
    sigma_prod = x_sigma * y_sigma
    core = n_weight * sigma_prod

    var = np.multiply.outer(weight_sq, core) + core
    std = var ** 0.5

    return std.T

def _global_spatialdm(x_mat,
                      y_mat,
                      weight,
                      seed,
                      n_perms,
                      mask_negatives,
                      verbose
                      ):
    # Get global r
    global_r = ((x_mat @ weight) * y_mat).sum(axis=1)

    # calc p-values
    if n_perms is None:
        global_pvals = None
    elif n_perms > 0:
        global_pvals = _global_permutation_pvals(x_mat=x_mat,
                                                 y_mat=y_mat,
                                                 weight=weight,
                                                 global_r=global_r,
                                                 n_perms=n_perms,
                                                 mask_negatives=mask_negatives,
                                                 seed=seed,
                                                 verbose=verbose
                                                 )
    elif n_perms==0:
        global_pvals = _global_zscore_pvals(weight=weight,
                                            global_r=global_r,
                                            mask_negatives=mask_negatives)

    return global_r, global_pvals


def _run_scores_pipeline(xy_stats, x_mat, y_mat, idx, local_fun, local_msk,
                         weight, mask_negatives, n_perms, seed, verbose):
    local_scores, local_pvals = _get_local_scores(x_mat=x_mat.T,
                                                  y_mat=y_mat.T,
                                                  local_fun=local_fun,
                                                  weight=weight,
                                                  seed=seed,
                                                  n_perms=n_perms,
                                                  mask_negatives=mask_negatives,
                                                  local_msk=local_msk,
                                                  verbose=verbose
                                                  )

    # global scores fun
    xy_stats = _get_global_scores(xy_stats=xy_stats,
                                  x_mat=x_mat,
                                  y_mat=y_mat,
                                  local_fun=local_fun,
                                  weight=weight,
                                  seed=seed,
                                  n_perms=n_perms,
                                  mask_negatives=mask_negatives,
                                  local_scores=local_scores,
                                  verbose=verbose
                                  )

    # convert to DataFrames
    local_scores = _local_to_dataframe(array=local_scores,
                                       idx=idx,
                                       columns=xy_stats['interaction'])

    return xy_stats, local_scores, local_pvals


def _get_local_scores(x_mat,
                      y_mat,
                      local_fun,
                      weight,
                      n_perms,
                      seed,
                      mask_negatives,
                      local_msk,
                      verbose
                      ):
    """
    Local Moran's Bivariate I as implemented in SpatialDM

    Returns
    -------
        Tupple of two 2D Numpy arrays of size (n_spots, n_xy),
         or in other words calculates local_I and local_pval for
         each interaction in `xy_dataframe` and each sample in mat
    """

    if local_fun.__name__ == '_local_morans':
        # TODO: probably just get rid of this
        # and leave it to the user
        x_mat = _zscore(x_mat, local=True, axis=0)
        y_mat = _zscore(y_mat, local=True, axis=0)

        # # NOTE: spatialdm do this, and also use .raw by default
        # x_mat = x_mat / np.max(x_mat, axis=0)
        # y_mat = y_mat / np.max(y_mat, axis=0)

    else:
        x_mat = x_mat.A
        y_mat = y_mat.A

    local_scores = local_fun(x_mat, y_mat, weight)

    if n_perms is None:
        local_pvals = None
    elif n_perms > 0:
        local_pvals = _local_permutation_pvals(x_mat=x_mat,
                                               y_mat=y_mat,
                                               weight=weight,
                                               local_truth=local_scores,
                                               local_fun=local_fun,
                                               n_perms=n_perms,
                                               seed=seed,
                                               mask_negatives=mask_negatives,
                                               local_msk=local_msk,
                                               verbose=verbose
                                               )
    elif n_perms == 0:
        local_pvals = _local_zscore_pvals(x_mat=x_mat,
                                          y_mat=y_mat,
                                          local_truth=local_scores,
                                          weight=weight,
                                          mask_negatives=mask_negatives,
                                          local_msk=local_msk
                                          )

    return local_scores, local_pvals


def _get_global_scores(xy_stats, x_mat, y_mat, local_fun, weight, mask_negatives,
                       n_perms, seed, local_scores, verbose):
    if local_fun.__name__ == "_local_morans":
        global_r, global_pvals = \
            _global_spatialdm(x_mat=_zscore(x_mat, local=False, axis=1),
                              y_mat=_zscore(y_mat, local=False, axis=1),
                              weight=weight,
                              seed=seed,
                              n_perms=n_perms,
                              mask_negatives=mask_negatives,
                              verbose=verbose
                              )
        xy_stats['global_r'] = global_r
        xy_stats['global_pvals'] = global_pvals
    else:
        # any other local score
        xy_stats.loc[:, ['global_mean', 'global_sd']] = np.vstack(
            [np.mean(local_scores, axis=1), np.std(local_scores, axis=1)]
            ).T

    return xy_stats


def _connectivity_to_weight(connectivity, local_fun):
    connectivity = csr_matrix(connectivity, dtype=np.float32)

    if local_fun.__name__ == "_local_morans":
        norm_factor = connectivity.shape[0] / connectivity.sum()
        connectivity = norm_factor * connectivity

        return csr_matrix(connectivity)

    # NOTE vectorized is faster with non-sparse, masked_scores won't work with sparse
    elif (connectivity.shape[0] < 5000) | local_fun.__name__.__contains__("masked"):
            return connectivity.A
    else:
        return csr_matrix(connectivity)


def _add_complexes_to_var(adata, entities, complex_sep='_'):
    """
    Generate an AnnData object with complexes appended as variables.
    """
    complexes = entities[Series(entities).str.contains(complex_sep)]

    X = None

    for comp in complexes:
        subunits = comp.split(complex_sep)

        # keep only complexes, the subunits of which are in var
        if all([subunit in adata.var.index for subunit in subunits]):
            adata.var.loc[comp, :] = None

            # create matrix for this complex
            new_array = csr_matrix(adata[:, subunits].X.min(axis=1))

            if X is None:
                X = new_array
            else:
                X = hstack((X, new_array))

    adata = AnnData(X=hstack((adata.X, X)),
                    obs=adata.obs, var=adata.var,
                    obsm=adata.obsm, obsp=adata.obsp)

    if not isspmatrix_csr(adata.X):
        adata.X = adata.X.tocsr()

    return adata



from __future__ import annotations

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, isspmatrix_csr
from anndata import AnnData
from mudata import MuData

from liana.method._pipe_utils._common import _get_props
from liana.method.sp._spatial_pipe import (
    _categorize,
    _rename_means,
    _run_scores_pipeline,
    _connectivity_to_weight,
    _add_complexes_to_var
    )
from liana.utils.obsm_to_adata import obsm_to_adata
from liana.utils.mdata_to_anndata import mdata_to_anndata
from liana.resource._select_resource import _handle_resource
from liana.method._pipe_utils import prep_check_adata, assert_covered
from liana.method.sp._bivariate_funs import _handle_functions, _bivariate_functions

from liana._logging import _logg
from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V


class SpatialBivariate():
    """ A class for bivariate local spatial metrics. """
    def __init__(self, x_name='x', y_name='y'):
        self.x_name = x_name
        self.y_name = y_name

    def _handle_return(self, data, stats, local_scores, key_added, x_added, inplace=False):
        if not inplace:
            return stats, local_scores

        data.uns[key_added] = stats

        if isinstance(data, MuData):
            data.mod[x_added] = local_scores
        else:
            data.obsm[x_added] = local_scores


    def _handle_connectivity(self, adata, connectivity_key):
        if connectivity_key not in adata.obsp.keys():
            raise ValueError(f'No connectivity matrix founds in mdata.obsp[{connectivity_key}]')
        connectivity = adata.obsp[connectivity_key]

        if not isspmatrix_csr(connectivity):
            connectivity = csr_matrix(connectivity, dtype=np.float32)

        return connectivity

    @d.dedent
    def __call__(self,
                 mdata: (MuData | AnnData),
                 x_mod: str,
                 y_mod: str,
                 function_name: str='cosine',
                 interactions: (None | list) = None,
                 resource: (None | pd.DataFrame) = None,
                 resource_name: (None | str) = None,
                 connectivity_key:str = K.connectivity_key,
                 mod_added: str = "local_scores",
                 key_added: str = 'global_res',
                 mask_negatives: bool = False,
                 add_categories: bool = False,
                 n_perms: int = None,
                 seed:int = V.seed,
                 nz_threshold:float = 0, # NOTE: do I rename this?
                 x_use_raw: bool = V.use_raw,
                 x_layer: (None | str) = V.layer,
                 x_transform: (bool | callable) = False,
                 y_use_raw: bool = V.use_raw,
                 y_layer: (None | str) = V.layer,
                 y_transform: (bool | callable) = False,
                 complex_sep: (None | str) = None,
                 xy_sep:str = V.lr_sep,
                 remove_self_interactions: bool = True,
                 inplace:bool = V.inplace,
                 verbose:bool = V.verbose,
                 ):
        """
        A method for bivariate local spatial metrics.

        Parameters
        ----------

        %(mdata)s
        %(x_mod)s
        %(y_mod)s
        %(function_name)s
        %(interactions)s
        %(resource)s
        %(resource_name)s
        %(connectivity_key)s
        mod_added: str
            Key in `mdata.mod` where the local scores are stored.
        %(key_added)s
        %(mask_negatives)s
        %(add_categories)s
        %(n_perms)s
        %(seed)s
        nz_threshold: float
            Minimum proportion of cells expressing the ligand and receptor.
        x_use_raw: bool
            Whether to use the raw counts for the x-mod.
        x_layer: str
            Layer to use for x-mod.
        x_transform: bool
            Function to transform the x-mod.
        y_use_raw: bool
            Whether to use the raw counts for y-mod.
        y_layer: str
            Layer to use for y-mod.
        y_transform: bool
            Function to transform the y-mod.
        complex_sep: str
            Separator to use for complex names.
        xy_sep: str
            Separator to use for interaction names.
        remove_self_interactions: bool
            Whether to remove self-interactions. `True` by default.
        %(inplace)s
        %(verbose)s

        Returns
        -------

        If `inplace` is `True`, the results are added to `mdata` and `None` is returned.
        If `inplace` is `False`, the results are returned.
        """

        if n_perms is not None:
            if not isinstance(n_perms, int) or n_perms < 0:
                raise ValueError("n_perms must be None, 0 for analytical or > 0 for permutation")

        local_fun = _handle_functions(function_name)

        resource = _handle_resource(interactions=interactions,
                                    resource=resource,
                                    resource_name=resource_name,
                                    x_name=self.x_name,
                                    y_name=self.y_name,
                                    verbose=verbose)

        if isinstance(mdata, MuData):
            adata = mdata_to_anndata(mdata,
                                     x_mod=x_mod,
                                     y_mod=y_mod,
                                     x_use_raw=x_use_raw,
                                     x_layer=x_layer,
                                     y_use_raw=y_use_raw,
                                     y_layer=y_layer,
                                     x_transform=x_transform,
                                     y_transform=y_transform,
                                     verbose=verbose,
                                     )
            use_raw = False
            layer = None
        elif isinstance(mdata, AnnData):
            adata = mdata
            use_raw = x_use_raw
            layer = x_layer
        else:
            raise ValueError("Invalid type, `adata/mdata` must be an AnnData/MuData object")

        adata = prep_check_adata(adata=adata,
                                 use_raw=use_raw,
                                 layer=layer,
                                 verbose=verbose,
                                 groupby=None,
                                 min_cells=None,
                                 complex_sep=complex_sep,
                                )


        connectivity = self._handle_connectivity(adata=adata, connectivity_key=connectivity_key)
        weight = _connectivity_to_weight(connectivity=connectivity, local_fun=local_fun)

        if complex_sep is not None:
            adata = _add_complexes_to_var(adata,
                                          np.union1d(resource[self.x_name].astype(str),
                                                     resource[self.y_name].astype(str)
                                                     ),
                                          complex_sep=complex_sep
                                          )

        # filter_resource
        resource = resource[(np.isin(resource[self.x_name], adata.var_names)) &
                            (np.isin(resource[self.y_name], adata.var_names))]

        # NOTE: Should I just get rid of remove_self_interactions?
        self_interactions = resource[self.x_name] == resource[self.y_name]
        if self_interactions.any() & remove_self_interactions:
            _logg(f"Removing {self_interactions.sum()} self-interactions", verbose=verbose)
            resource = resource[~self_interactions]

        # get entities
        entities = np.union1d(np.unique(resource[self.x_name]),
                                np.unique(resource[self.y_name]))
        # Check overlap between resource and adata TODO check if this works
        assert_covered(entities, adata.var_names, verbose=verbose)

        # Filter to only include the relevant features
        adata = adata[:, np.intersect1d(entities, adata.var.index)]

        xy_stats = pd.DataFrame({'means': adata.X.mean(axis=0).A.flatten(),
                                 'props': _get_props(adata.X)},
                                index=adata.var_names
                                ).reset_index().rename(columns={'index': 'gene'})
        # join global stats to LRs from resource
        xy_stats = resource.merge(_rename_means(xy_stats, entity=self.x_name)).merge(
            _rename_means(xy_stats, entity=self.y_name))

        # TODO: nz_threshold to nz_prop? For consistency with other methods
        # filter according to props
        xy_stats = xy_stats[(xy_stats[f'{self.x_name}_props'] >= nz_threshold) &
                            (xy_stats[f'{self.y_name}_props'] >= nz_threshold)]
        # create interaction column
        xy_stats['interaction'] = xy_stats[self.x_name] + xy_sep + xy_stats[self.y_name]

        x_mat = adata[:, xy_stats[self.x_name]].X.T
        y_mat = adata[:, xy_stats[self.y_name]].X.T

        # reorder columns, NOTE: why?
        xy_stats = xy_stats.reindex(columns=sorted(xy_stats.columns))

        if add_categories or mask_negatives:
            local_cats = _categorize(x_mat=x_mat,
                                     y_mat=y_mat,
                                     weight=weight,
                                     idx=mdata.obs.index,
                                     columns=xy_stats['interaction'],
                                     )
            local_msk = local_cats != 0
        else:
            local_cats = None
            local_msk = None

        # get local scores
        xy_stats, local_scores, local_pvals = \
            _run_scores_pipeline(xy_stats=xy_stats,
                                 x_mat=x_mat,
                                 y_mat=y_mat,
                                 idx=mdata.obs.index,
                                 local_fun=local_fun,
                                 weight=weight,
                                 seed=seed,
                                 n_perms=n_perms,
                                 mask_negatives=mask_negatives,
                                 local_msk=local_msk,
                                 verbose=verbose,
                                 )
        local_scores = obsm_to_adata(adata=mdata, df=local_scores, obsm_key=None, _uns=mdata.uns)
        local_scores.uns[key_added] = xy_stats

        if mask_negatives:
            local_scores.X = local_scores.X * local_msk.T
        if local_cats is not None:
            local_scores.layers['cats'] = csr_matrix(local_cats.T)
        if local_pvals is not None:
            local_scores.layers['pvals'] = csr_matrix(local_pvals.T)

        return self._handle_return(mdata, xy_stats, local_scores, key_added, mod_added, inplace)

    def show_functions(self):
        """
        Print information about all bivariate local metrics.
        """
        funs = dict()
        for function in _bivariate_functions:
            funs[function.name] = {
                "metadata":function.metadata,
                "reference":function.reference,
                }

        return pd.DataFrame(funs).T.reset_index().rename(columns={"index":"name"})

bivar = SpatialBivariate()



class DefaultValues():
    """Default Values"""
    logbase = 2.718281828459045
    min_cells = 5
    expr_prop = 0.1
    n_perms = 1000
    seed = 1337
    de_method = 't-test'
    resource_name = 'consensus'
    resource = None
    interactions = None
    layer = None
    use_raw = True
    verbose = False
    return_all_lrs = False
    supp_columns = None
    inplace = True

    return_fig = True
    cmap = 'viridis'

    lr_sep = '^'
    complex_sep = "_"

    def inverse_fun(x):
        return 1 - x

class Keys():
    """Keys related to AnnData"""
    uns_key = 'liana_res'
    spatial_key = 'spatial'
    connectivity_key = f'{spatial_key}_connectivities'
    target_metrics = 'target_metrics'
    interactions = 'interactions'

class PrimaryColumns():
    source = 'source'
    target = 'target'
    ligand = 'ligand'
    receptor = 'receptor'
    ligand_complex = 'ligand_complex'
    receptor_complex = 'receptor_complex'
    primary = [source, target, ligand_complex, receptor_complex]
    complete = primary + [ligand, receptor]

class CommonColumns():
    ligand_means = 'ligand_means'
    receptor_means = 'receptor_means'
    ligand_props = 'ligand_props'
    receptor_props = 'receptor_props'
    ligand_pvals = 'ligand_pvals'
    receptor_pvals = 'receptor_pvals'

class MethodColumns():
    ligand_means_sums = 'ligand_means_sums'
    receptor_means_sums = 'receptor_means_sums'
    ligand_zscores = 'ligand_zscores'
    receptor_zscores = 'receptor_zscores'
    ligand_logfc = 'ligand_logfc'
    receptor_logfc = 'receptor_logfc'
    ligand_trimean = 'ligand_trimean'
    receptor_trimean = 'receptor_trimean'
    mat_mean = 'mat_mean'
    mat_max = 'mat_max'

    @classmethod
    def get_all_values(cls):
        return [value for name, value in cls.__dict__.items()
                if not name.startswith('__') and isinstance(value, str)]

class InternalValues():
    lrs_to_keep = 'lrs_to_keep'
    prop_min = 'prop_min'
    label = '@label'



from __future__ import annotations

import numpy as np
import pandas as pd
from anndata import AnnData
from liana._docs import d

@d.dedent
def obsm_to_adata(adata: AnnData,
                  obsm_key: str,
                  df: (pd.DataFrame | None) = None,
                  _uns: (pd.DataFrame | None) =None,
                  _obsm: (pd.DataFrame | None)=None
                  ):
    """
    Extracts a dataframe from adata.obsm and returns a new AnnData object with the values stored in X.

    Parameters
    ----------

    %(adata)s
    obsm_key
        `.osbm` key to extract.
    df
        Dataframe with stats per cell/spot. If None, it will be extracted from adata.obsm[obsm_key].
    _uns
        Dictionary with uns data. If None, it will be extracted from adata.uns.
    _obsm
        Dictionary with obsm data. If None, it will be extracted from adata.obsm.

    Returns
    -------
    An AnnData object with the values stored in X.
    """

    if df is None:
        df = adata.obsm[obsm_key]

    obs = adata.obs

    if _uns is None:
        uns = adata.uns
    else:
        uns = _uns

    if _obsm is None:
        obsm = adata.obsm
    else:
        obsm = _obsm

    obsp = adata.obsp

    var = pd.DataFrame(index = df.columns)
    X = np.array(df, dtype=np.float32)

    return AnnData(X=X, obs=obs, var=var, uns=uns, obsm=obsm, obsp=obsp)



from liana.multi.df_to_lr import df_to_lr
from liana.multi.to_tensor_c2c import to_tensor_c2c
from liana.multi.to_mudata import adata_to_views, lrs_to_views, filter_view_markers
from liana.multi._nmf import nmf, estimate_elbow



import anndata as an
from liana.method._pipe_utils._pre import _choose_mtx_rep
from liana._docs import d

@d.dedent
def mdata_to_anndata(mdata,
                     x_mod, y_mod,
                     x_layer=None, y_layer=None,
                     x_use_raw=False, y_use_raw=False,
                     x_transform=None,
                     y_transform=None,
                     verbose=True
                     ):
    """
    Convert a MultiData object to an AnnData object.

    Parameters
    ----------
    mdata
        MuData object.
    x_mod
        Name of the modality to be used as x.
    y_mod
        Name of the modality to be used as y.
    x_layer
        Layer to be used for modality x.
    y_layer
        Layer to be used for modality y.
    x_use_raw
        Whether to use raw counts for modality x.
    y_use_raw
        Whether to use raw counts for modality y.
    x_transform
        Transformation function to be applied to modality x.
    y_transform
        Transformation function to be applied to modality y.
    %(verbose)s

    Returns
    -------
    An AnnData object with the two modalities concatenated.
    Information related to observations (obs, obsp, obsm) and `.uns` are copied from the original MuData object.
    """

    if x_mod is None or y_mod is None:
        raise ValueError("Both `x_mod` and `y_mod` must be provided!")

    xdata = _handle_mod(mdata, x_mod, x_use_raw, x_layer, x_transform, verbose)
    ydata = _handle_mod(mdata, y_mod, y_use_raw, y_layer, y_transform, verbose)

    adata = an.concat([xdata, ydata], axis=1, label='modality')

    adata.obs = mdata.obs.copy()
    adata.obsp = mdata.obsp.copy()
    adata.obsm = mdata.obsm.copy()
    adata.uns = mdata.uns.copy()

    return adata

def _handle_mod(mdata, mod, use_raw, layer, transform, verbose):
    if mod not in mdata.mod.keys():
        raise ValueError(f'`{mod}` is not in the mdata!')

    # NOTE, maybe instead of copying I can just create a minimal AnnData?
    md = mdata.mod[mod].copy()
    md.X = _choose_mtx_rep(md, use_raw = use_raw, layer = layer, verbose=verbose)

    if transform:
        if verbose:
            print(f'Transforming {mod} using {transform.__name__}')
        md.X = transform(md.X)
    return md



import numpy as np
from liana.plotting import dotplot, dotplot_by_sample, tileplot
from liana.testing import sample_lrs
from liana.testing import generate_toy_spatial

liana_res = sample_lrs()

def test_dotplot_order():
    my_p = dotplot(liana_res=liana_res,
                   size='specificity_rank',
                   colour='magnitude',
                   top_n=20,
                   orderby='specificity_rank',
                   orderby_ascending=False,
                   target_labels=["A", "B", "C"]
                   )
    assert my_p is not None
    assert 'interaction' in my_p.data.columns
    np.testing.assert_equal(np.unique(my_p.data.interaction).shape, (20,))
    set(my_p.data.target)
    assert {'A', 'B', 'C'} == set(my_p.data.target)


def test_doplot_filter():
    my_p2 = dotplot(liana_res=liana_res,
                    size='specificity_rank',
                    colour='magnitude',
                    filterby='specificity_rank',
                    filter_lambda=lambda x: x > 0.95,
                    inverse_colour=True,
                    source_labels=["A"]
                    )
    assert my_p2 is not None
    # we force this, but not intended all interactions
    # to be only 0.95, but rather for an interaction to get
    # plotted, in at least one cell type pair it should be > 0.95
    assert all(my_p2.data['specificity_rank'] > 0.95) is True


def test_dotplot_bysample():
    liana_res = sample_lrs(by_sample=True)
    my_p3 = dotplot_by_sample(liana_res=liana_res,
                              size='specificity_rank',
                              colour='magnitude',
                              target_labels='E',
                              sample_key='sample')
    assert my_p3 is not None
    assert 'interaction' in my_p3.data.columns
    assert 'sample' in my_p3.data.columns
    assert 'B' not in my_p3.data['target']


def test_tileplot():
    my_p2 = tileplot(liana_res = liana_res,
                     # NOTE: fill & label need to exist for both
                     # ligand_ and receptor_ columns
                     fill='means',
                     label='pvals',
                     label_fun=lambda x: f'{x:.2f}',
                     top_n=10,
                     orderby='specificity_rank',
                     orderby_ascending=True
                     )
    assert my_p2 is not None
    assert isinstance(my_p2.data['pvals'].values[0], str)


def test_proximity_plot():
    from liana.plotting import connectivity

    adata = generate_toy_spatial()
    my_p4 = connectivity(adata=adata, idx=0)
    assert my_p4 is not None



from liana.utils.obsm_to_adata import obsm_to_adata
from liana.utils.mdata_to_anndata import mdata_to_anndata
from liana.utils.transform import zi_minmax, neg_to_zero
from liana.utils.spatial_neighbors import spatial_neighbors
from liana.utils._getters import get_factor_scores, get_variable_loadings
from liana.utils.query_bandwidth import query_bandwidth
